/* $Id: comm.c,v 1.43 2007-05-29 20:32:05 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>

#include <assert.h>
#include <err.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <libgen.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <curl/curl.h>

#include "nlpcrawl.h"
#include "comm.h"


/*
 * Sets the timeout for DNS queries.  Officially, this component
 * shouldn't be doing any DNS queries; they should have been handled by
 * the nlpc-resolv utility.  This has yet to be confirmed.  By being set
 * to -1, we disable DNS timeouts.
 */
#define COMM_TIMEO_DNS	(-1)

/*
 * Seconds before the initial connection event times out.  This doesn't
 * prevent sites from edging in just before the timeout and indefinitely
 * extending the flow of information after (which would DOS the sytsem).
 */
#define COMM_TIMEO_CONN	(60)



static struct opt_t
{
	CURLSH		*share;		/* Shared handle data.       */
	CURLM		*cookie;	/* Multi handle data.        */
	void		*lang;		/* Language dict (or NULL).  */
	void		*enc;		/* Encoding dict (or NULL).  */
	void		*queue;		/* Write queue.              */
	int		ascend;		/* Whether to ascend.        */
} opts;


/*
 * Header containers for persistent, un-changing headers: language
 * (Accept-Language), character set (Accept-Charset) and User-Agent.
 */
static char		head_lang[COMM_HEAD_SZ];
static char		head_charset[COMM_HEAD_SZ];
static char		head_agent[COMM_HEAD_SZ];



static void
comm_getdate(char *source, uint64_t *dest)
{
	time_t		tm;

	assert(NULL != source);
	assert(NULL != dest);
	if (-1 == curl_getdate(source, &tm))
		diex("%s: fail: curl_getdate: %s", __func__, source);
	*dest = (uint64_t)tm;
}


static void
comm_stats_double(struct comm_t *p, CURLINFO opt, double *val)
{
	CURLcode	ret;
	CURL		*curl;

	assert(NULL != val);
	assert(NULL != p);
	assert(NULL != p->handle);

	curl = (CURL *)p->handle;
	if (CURLE_OK == (ret = curl_easy_getinfo(curl, opt, val)))
		return;

	diex("%s: fail: %s, %s", __func__,
			curl_easy_strerror(ret), p->pretty);
	/*NOTREACHED*/
}


static void
comm_stats_fill(struct comm_t *p, struct comm_stats *s)
{
	double		d;

	assert(NULL != p);
	assert(NULL != s);

	comm_stats_double(p, CURLINFO_TOTAL_TIME, &d);
	s->ttime = (uint64_t)d;
	comm_stats_double(p, CURLINFO_NAMELOOKUP_TIME, &d);
	s->nl_time = (uint64_t)d;
	comm_stats_double(p, CURLINFO_CONNECT_TIME, &d);
	s->ct_time = (uint64_t)d;
	comm_stats_double(p, CURLINFO_PRETRANSFER_TIME, &d);
	s->pt_time = (uint64_t)d;
	comm_stats_double(p, CURLINFO_STARTTRANSFER_TIME, &d);
	s->st_time = (uint64_t)d;
	comm_stats_double(p, CURLINFO_SIZE_DOWNLOAD, &d);
	s->dl_size = (uint64_t)d;
}


static void
comm_opt_long(struct comm_t *p, CURLoption opt, long val)
{
	CURLcode	ret;

	assert(NULL != p);
	assert(NULL != p->handle);
	ret = curl_easy_setopt((CURL *)p->handle, opt, val);
	if (CURLE_OK == ret)
		return;

	diex("%s: fail: %s, %s", __func__,
			curl_easy_strerror(ret), p->pretty);
	/*NOTREACHED*/
}


static void
comm_opt_void(struct comm_t *p, CURLoption opt, void *val)
{
	CURLcode	ret;

	assert(NULL != p);
	assert(NULL != p->handle);
	assert(NULL != val);
	ret = curl_easy_setopt((CURL *)p->handle, opt, val);
	if (CURLE_OK == ret)
		return;

	diex("%s: fail: %s, %s", __func__,
			curl_easy_strerror(ret), p->pretty);
	/*NOTREACHED*/
}


static void
comm_hdr_add(struct comm_t *p, char field[COMM_HEAD_SZ], 
		const char *key, const char *val, size_t sz)
{
	assert(0 == field[0]);
	assert(NULL != p);
	assert(NULL != key);
	assert(NULL != val);

	strncpy(field, key, sz - 1);
	assert(0 == field[sz - 1]);

	strncat(field, val, sz - strlen(field) - 1);
	assert(0 == field[sz - 1]);

	dddbg("%s: header: [%s]", __func__, field);

	p->head.handle = curl_slist_append(p->head.handle, field);
	assert(NULL != p->head.handle);

}


static void
comm_hdr_init(char field[COMM_HEAD_SZ], const char *key, char *val)
{
	char		*t, *tt;

	assert(0 == field[0]);
	assert(NULL != key);
	assert(NULL != val);

	if (COMM_HEAD_SZ < strlen(val) + strlen(key) + 10)
		diex("%s: fail: header overrun: %s", __func__, key);

	while (NULL != (t = strsep(&val, " "))) { 
		if (0 == strlen(t))
			continue;
		for (tt = t; 0 != *tt; tt++) 
			if (isalpha(*tt))
				*tt = tolower(*tt);
		strncat(field, t, COMM_HEAD_SZ - strlen(field) - 1);
		strncat(field, ",", COMM_HEAD_SZ - strlen(field) - 1);
	} 
}


static int
comm_content_lang(struct http_header *h, struct comm_t *c)
{
	struct nlpcdbr	*r;
	char		*lang;

	assert(NULL != h);
	assert(NULL != c);
	assert(HTTP_CONTENT_LANG == h->type);
	assert(COMM_RECORD == c->type);
	assert(NULL != c->data.pair);

	r = &c->data.pair->record;
	lang = h->data.clang;

	memset(r->lang, 0, NLPCR_CHARSETLEN);
	if (NULL != lang) {
		strncpy(r->lang, lang, NLPCR_LANGLEN);
		if (0 == lang[0])
			return(1);
	} else
		return(1);

	if (0 != r->lang[NLPCR_LANGLEN - 1]) {
		ddbg("%s: warn: language overrun: [%s], %s", 
				__func__, lang, c->pretty);
		r->state = NLPCDBR_UNREACH;
		r->code = NLPCDBR_SYNTAX;
		return(0);
	} 

	if ( ! dict_has(&opts.lang, lang, strlen(lang))) {
		ddbg("%s: warn: non-matching language: [%s], %s", 
				__func__, lang, c->pretty);
		r->state = NLPCDBR_UNREACH;
		r->code = NLPCDBR_LANG;
		return(0);
	}

	return(1);
}


static int
comm_content_type(struct http_header *h, struct comm_t *c)
{
	struct nlpcdbr	*r;
	char		*ctype;
	uint64_t	media;

	assert(NULL != h);
	assert(NULL != c);
	assert(HTTP_CONTENT_TYPE == h->type);
	assert(COMM_RECORD == c->type);
	assert(NULL != c->data.pair);

	r = &c->data.pair->record;
	ctype = h->data.ctype.charset;
	media = h->data.ctype.media;

	if (NLPCDBR_TYPE_NONE == (r->type = media)) {
		ddbg("%s: warn: unknown media: %s", __func__, 
				c->pretty);
		r->state = NLPCDBR_UNREACH;
		r->code = NLPCDBR_CTYPE;
		return(0);
	}

	memset(r->charset, 0, NLPCR_CHARSETLEN);
	if (NULL != ctype) {
		strncpy(r->charset, ctype, NLPCR_CHARSETLEN);
		if (0 == ctype[0])
			return(1);
	} else
		return(1);

	if (0 != r->charset[NLPCR_CHARSETLEN - 1]) {
		ddbg("%s: warn: charset overrun: [%s], %s", 
				__func__, ctype, c->pretty);
		r->state = NLPCDBR_UNREACH;
		r->code = NLPCDBR_SYNTAX;
		return(0);
	}

	if ( ! dict_has(&opts.enc, ctype, strlen(ctype))) {
		ddbg("%s: warn: non-matching charset: [%s], %s", 
				__func__, ctype, c->pretty);
		r->state = NLPCDBR_UNREACH;
		r->code = NLPCDBR_CHARSET;
		return(0);
	}

	return(1);
}


static void
comm_enqueue(const struct uri *addr)
{
	struct nlpcdbrp	*n;
	uint64_t	flags;

	assert(NULL != addr);

	flags = SQUEUE_NOTIFY | SQUEUE_WAIT;
	calloc_s(n, 1, sizeof(struct nlpcdbrp));
	nlpcr_new(addr, n);
	squeue_add(opts.queue, n, flags);
}


size_t
rep_head(void *ptr, size_t smem, size_t nmem, void *arg)
{
	struct http_header h;
	struct comm_t 	*c;

	assert(NULL != arg);
	assert(NULL != ptr);
	assert(1 == smem);

	c = (struct comm_t *)arg;
	assert(COMM_REP == c->type);

	if (0 == c->status) {
		if ( ! http_status((char *)ptr, nmem, &c->status)) 
			return(nmem + 1);
		switch(c->status) {
		case (200):
			break;
		default:
			return(nmem + 1);
		}
		return(nmem);
	}

	if ( ! http_parse((char *)ptr, nmem, &h))
		return(nmem);

	switch (h.type) {
	case (HTTP_EXPIRES):
		comm_getdate(h.data.expires, &c->data.rep.expires);
		break;
	default:
		break;
	}

	return(nmem);
}


size_t
rep_data(void *ptr, size_t smem, size_t nmem, void *arg)
{
	ssize_t		rc;
	struct comm_t	*p;

	assert(NULL != arg);
	assert(NULL != ptr);
	assert(1 == smem);

	p = (struct comm_t *)arg;
	assert(COMM_REP == p->type);

	if (-1 == (rc = write(p->fd, ptr, nmem)))
		die("write: %s", p->pretty);
	if (rc != (ssize_t)nmem) 
		diex("write (underrun): %s", p->pretty);
	return(nmem);
}


size_t
comm_head(void *ptr, size_t smem, size_t nmem, void *arg)
{
	struct http_header h;
	struct comm_t 	*co;
	struct nlpcdbr	*r;

	assert(NULL != arg);
	assert(NULL != ptr);
	assert(1 == smem);

	co = (struct comm_t *)arg;

	assert(COMM_RECORD == co->type);
	assert(NULL != co->data.pair);

	r = &co->data.pair->record;

#define mod(st, cd) do {                                              \
	co->data.pair->record.state = (st);                           \
	co->data.pair->record.code = (cd); } while(0)

#define mod_ret(st, cd) do {                                          \
	mod((st), (cd));                                              \
	return(nmem + 1);                                             \
	/* NOTREACHED */ } while(0)

	if (0 == co->status) {
		if ( ! http_status((char *)ptr, nmem, &co->status)) 
			mod_ret(NLPCDBR_UNREACH, NLPCDBR_SYNTAX);
		else if (200 != co->status)
			ddbg("%s: bad http status-code: %u, %s", 
				__func__, co->status, co->pretty);

		switch(co->status) {
		case (200):
			break;
		case (300):
			/* FALLTHROUGH */
		case (302):
			/* FALLTHROUGH */
		case (303):
			/* FALLTHROUGH */
		case (307):
			mod(NLPCDBR_UNREACH, NLPCDBR_MOVED);
			break;
		case (301):
			mod(NLPCDBR_DEAD, NLPCDBR_MOVED);
			break;
		case (304):
			mod_ret(NLPCDBR_FRESH, NLPCDBR_OK);
			/* NOTREACHED */
		default:
			mod_ret(NLPCDBR_UNREACH, NLPCDBR_SERVER);
			/* NOTREACHED */
		}
		return(nmem);
	}

	if ( ! http_parse((char *)ptr, nmem, &h))
		return(nmem);

	switch (h.type) {
	case (HTTP_EXPIRES):
		comm_getdate(h.data.expires, &r->expires);
		break;
	case (HTTP_PRAGMA):
		if (HTTP_PRAGMA_NOCACHE == h.data.pragma.pragma_type)
			if (h.data.pragma.data.nocache)
				r->expires = UINT64_MAX;
		break;
	case (HTTP_CONTENT_TYPE):
		if ( ! comm_content_type(&h, co))
			return(nmem + 1);
		break;
	case (HTTP_CONTENT_LANG):
		if ( ! comm_content_lang(&h, co))
			return(nmem + 1);
		break;
	case (HTTP_LOCATION):
		if ( ! opts.ascend) 
			comm_enqueue(&h.data.location);
		else while (uri_ascend(&h.data.location)) 
			comm_enqueue(&h.data.location);
		return(nmem + 1);
	default:
		break;
	}

	return(nmem);
}


void
comm_hdr_lang(char *val)
{
	comm_hdr_init(head_lang, "Accept-Language", val);
}


void
comm_hdr_charset(char *val)
{
	comm_hdr_init(head_charset, "Accept-Content", val);
}


void
comm_hdr_agent(const char *val)
{
	assert(NULL != val);
	assert(0 == head_agent[0]);

	if (COMM_HEAD_SZ < strlen(val) + strlen("User-Agent: ") + 1)
		diex("%s: fail: header overrun: %s", __func__, val);

	strncat(head_agent, val, COMM_HEAD_SZ - 
			strlen(head_agent) - 1);
	assert(0 == head_agent[COMM_HEAD_SZ - 1]);
}


void
comm_free(struct comm_t *p)
{
	assert(NULL != p);
	assert(NULL != p->handle);

	curl_easy_cleanup((CURL *)p->handle);
	p->handle = NULL;
}


void
rep_close(struct comm_t *p, struct comm_stats *s)
{
	CURLMcode	ret;
	CURL		*curl;

	assert(NULL != p);
	assert(NULL != p->handle);
	assert(NULL != opts.cookie);
	assert(NULL == p->hash);
	assert(-1 != p->fd);
	assert(COMM_REP == p->type);

	curl = (CURL *)p->handle;

	if (-1 == close(p->fd))
		die("%s: fail: close: %s", __func__, p->pretty);
	p->fd = -1;

	assert(NULL != p->head.handle);
	curl_slist_free_all((struct curl_slist *)p->head.handle);
	
	if (NULL != s)
		comm_stats_fill(p, s);

	ret = curl_multi_remove_handle(opts.cookie, curl);
	if (CURLM_OK == ret)
		return;

	diex("%s: fail: %s, %s", __func__,
			curl_multi_strerror(ret), p->pretty);
	/* NOTREACHED */
}


void
comm_close(struct comm_t *p, struct comm_stats *s)
{
	CURLMcode	ret;
	CURL		*curl;

	assert(NULL != p);
	assert(NULL != p->handle);
	assert(NULL != opts.cookie);
	assert(NULL != p->hash);
	assert(-1 != p->fd);
	assert(COMM_RECORD == p->type);
	assert(NULL != p->data.pair);

	curl = (CURL *)p->handle;

	if (-1 == close(p->fd))
		die("%s: fail: close: %s", __func__, p->pretty);
	p->fd = -1;

	assert(NULL != p->head.handle);
	curl_slist_free_all((struct curl_slist *)p->head.handle);
	
	md5_hash_close(&p->hash, p->data.pair->record.data);

	if (NULL != s)
		comm_stats_fill(p, s);

	ret = curl_multi_remove_handle(opts.cookie, curl);
	if (CURLM_OK == ret)
		return;

	diex("%s: fail: %s, %s", __func__,
			curl_multi_strerror(ret), p->pretty);
	/*NOTREACHED*/
}


void
comm_new(struct comm_t *p)
{
	assert(NULL != p);
	p->fd = -1;
	if (NULL == (p->handle = curl_easy_init()))
		diex("fail: %s: curl_easy_init", __func__);
}


size_t
comm_data(void *ptr, size_t smem, size_t nmem, void *arg)
{
	ssize_t		rc;
	struct comm_t	*p;

	assert(NULL != arg);
	assert(NULL != ptr);
	assert(1 == smem);

	p = (struct comm_t *)arg;
	assert(COMM_RECORD == p->type);
	assert(NULL != p->data.pair);
	assert(NULL != p->hash);

	md5_hash_update((const char *)ptr, nmem, p->hash);

	if (-1 == (rc = write(p->fd, ptr, nmem)))
		die("write: %s", p->pretty);
	if (rc != (ssize_t)nmem) 
		diex("write (underrun): %s", p->pretty);

	p->data.pair->record.size += nmem;
	return(nmem);
}


void
rep_open(struct comm_t *p, const char *prefix, struct nlpcdbrp *pair)
{
	CURLMcode	ret;
	CURL		*curl;
	char		path[MAXPATHLEN], dir[MAXPATHLEN], *dp;

	assert(NULL != pair);
	assert(NULL != prefix);
	assert(NULL != p);
	assert(NULL != p->handle);
	assert(NULL == p->hash);
	assert(NULL == p->data.pair);
	assert(0 == p->status);
	assert(-1 == p->fd);
	assert(NULL == p->head.handle);

	p->type = COMM_REP;
	memcpy(p->data.rep.ip, pair->record.ip, NLPCR_IP_WIDTH);
	memcpy(p->data.rep.auth, pair->record.uri.auth, 
			NLPCR_URI_AUTHSZ);

	curl = (CURL *)p->handle;

	assert(0 == p->uri[0]);
	strncat(p->uri, "http://", NLPCR_URI_MAXSZ);
	ip_tostring_r(p->data.rep.ip, p->uri + strlen(p->uri));
	strncat(p->uri + strlen(p->uri), "/robots.txt",
			NLPCR_URI_MAXSZ - strlen(p->uri) - 1);

	strncat(p->pretty, "[http][", NLPCR_URI_FMTSZ);
	strncat(p->pretty, p->data.rep.auth, NLPCR_URI_FMTSZ -
			strlen(p->pretty) - 1);
	strncat(p->pretty + strlen(p->pretty), "][/robots.txt]",
			NLPCR_URI_MAXSZ - strlen(p->uri) - 1);

	comm_hdr_add(p, p->head.agent, "User-Agent: ", 
			head_agent, COMM_HEAD_SZ);
	assert(0 == p->head.charset[0]);
	if (0 != head_lang[0])
		comm_hdr_add(p, p->head.charset, "Accept-Content: ", 
				head_charset, COMM_HEAD_SZ);

	assert(0 == p->head.host[0]);
	comm_hdr_add(p, p->head.host, "Host: ", 
			pair->record.uri.auth,
			COMM_HEAD_SZ + NLPCR_URI_AUTHSZ);

	assert(NULL != opts.share);

	comm_opt_void(p, CURLOPT_URL, p->uri);
	comm_opt_void(p, CURLOPT_WRITEDATA, p);
	comm_opt_void(p, CURLOPT_HEADERDATA, p);
	comm_opt_void(p, CURLOPT_HEADERFUNCTION, rep_head);
	comm_opt_void(p, CURLOPT_WRITEFUNCTION, rep_data);
	comm_opt_void(p, CURLOPT_HTTPHEADER, p->head.handle);
	comm_opt_void(p, CURLOPT_SHARE, opts.share);

	comm_opt_long(p, CURLOPT_MAXCONNECTS, 30);
	comm_opt_long(p, CURLOPT_FOLLOWLOCATION, 0);
	comm_opt_long(p, CURLOPT_TIMEOUT, COMM_TIMEO_CONN);
	comm_opt_long(p, CURLOPT_NOSIGNAL, 1);
	comm_opt_long(p, CURLOPT_DNS_CACHE_TIMEOUT, COMM_TIMEO_DNS);

	assert(NULL != opts.cookie);
	assert(NULL != curl);

	ret = curl_multi_add_handle(opts.cookie, curl);
	if (CURLM_OK != ret)
		diex("%s: fail: %s, %s", __func__,
				curl_multi_strerror(ret), p->pretty);

	memset(path, 0, MAXPATHLEN);
	memset(dir, 0, MAXPATHLEN);

	strncpy(path, prefix, MAXPATHLEN);
	if ('/' != prefix[strlen(prefix) - 2])
		strncat(path, "/", MAXPATHLEN - strlen(prefix) - 1);
	assert(strlen(path) + NLPCR_IP_STRING < MAXPATHLEN - 1);
	ip_tostring_r(p->data.rep.ip, path + strlen(path));

	memcpy(dir, path, MAXPATHLEN);

	if (NULL == (dp = dirname(dir)))
		die("%s: fail: dirname: %s", __func__, path);
	if (mkdir_p(dp, 0755))
		die("%s: fail: mkdir_p: %s", __func__, dp);

	p->fd = open(path, O_RDWR | O_TRUNC | O_CREAT, 0644);
	if (-1 == p->fd) 
		die("%s: fail: open: %s", __func__, path);
}


void
comm_open(struct comm_t *p, const char *prefix, struct nlpcdbrp *pair)
{
	CURLMcode	ret;
	CURL		*curl;
	struct nlpcdbr	*r;
	char		path[MAXPATHLEN], dir[MAXPATHLEN], *dp, *t;
	char		dt[32];
	time_t		tm;

	assert(NULL != prefix);
	assert(NULL != pair);
	assert(NULL != p);
	assert(NULL != p->handle);
	assert(NULL == p->hash);
	assert(NULL == p->data.pair);
	assert(0 == p->status);
	assert(-1 == p->fd);
	assert(NULL == p->head.handle);

	p->type = COMM_RECORD;
	p->data.pair = pair;
	r = &p->data.pair->record;
	curl = (CURL *)p->handle;

	assert(0 == p->uri[0]);
	uri_ipmerge_r(&r->uri, r->ip, p->uri);

	assert(0 == p->pretty[0]);
	uri_topstring_r(&r->uri, p->pretty);

	if (0 != r->ftime) {
		tm = (time_t)r->ftime;
		memcpy(dt, ctime(&tm), 32);
		dt[strlen(dt) - 1] = 0;
		comm_hdr_add(p, p->head.date, "If-Modified-Since: ", 
				dt, COMM_HEAD_SZ);
	}

	comm_hdr_add(p, p->head.agent, "User-Agent: ", 
			head_agent, COMM_HEAD_SZ);

	assert(0 == p->head.lang[0]);
	if (0 != head_lang[0])
		comm_hdr_add(p, p->head.lang, "Accept-Language: ", 
				head_lang, COMM_HEAD_SZ);

	assert(0 == p->head.charset[0]);
	if (0 != head_lang[0])
		comm_hdr_add(p, p->head.charset, "Accept-Content: ", 
				head_charset, COMM_HEAD_SZ);

	assert(0 == p->head.host[0]);
	comm_hdr_add(p, p->head.host, "Host: ", r->uri.auth, 
			COMM_HEAD_SZ + NLPCR_URI_AUTHSZ);

	assert(NULL != opts.share);

	comm_opt_void(p, CURLOPT_URL, p->uri);
	comm_opt_void(p, CURLOPT_WRITEDATA, p);
	comm_opt_void(p, CURLOPT_HEADERDATA, p);
	comm_opt_void(p, CURLOPT_HEADERFUNCTION, comm_head);
	comm_opt_void(p, CURLOPT_WRITEFUNCTION, comm_data);
	comm_opt_void(p, CURLOPT_HTTPHEADER, p->head.handle);
	comm_opt_void(p, CURLOPT_SHARE, opts.share);

	comm_opt_long(p, CURLOPT_MAXCONNECTS, 30);
	comm_opt_long(p, CURLOPT_FOLLOWLOCATION, 0);
	comm_opt_long(p, CURLOPT_TIMEOUT, COMM_TIMEO_CONN);
	comm_opt_long(p, CURLOPT_NOSIGNAL, 1);
	comm_opt_long(p, CURLOPT_DNS_CACHE_TIMEOUT, COMM_TIMEO_DNS);

	assert(NULL != opts.cookie);
	assert(NULL != curl);

	ret = curl_multi_add_handle(opts.cookie, curl);
	if (CURLM_OK != ret)
		diex("%s: fail: %s, %s", __func__,
				curl_multi_strerror(ret), p->pretty);

	md5_hash_open(&p->hash);

	memset(path, 0, MAXPATHLEN);
	memset(dir, 0, MAXPATHLEN);

	t = nlpcdbr_hash2path(prefix, pair->hash);

	memcpy(path, t, MAXPATHLEN);
	memcpy(dir, path, MAXPATHLEN);

	if (NULL == (dp = dirname(dir)))
		die("%s: fail: dirname: %s", __func__, path);
	if (mkdir_p(dp, 0755))
		die("%s: fail: mkdir_p: %s", __func__, dp);

	p->fd = open(path, O_WRONLY | O_TRUNC | O_CREAT, 0644);
	if (-1 == p->fd) 
		die("%s: fail: open: %s", __func__, path);
}


void
comm_init(void *lang, void *enc, void *queue, int ascend)
{
	CURLcode 	cc;
	CURLSHcode	cs;
	CURLMcode	cm;

	assert(NULL != queue);

	opts.lang = lang;
	opts.enc = enc;
	opts.queue = queue;
	opts.ascend = ascend;

	cc = curl_global_init(CURL_GLOBAL_ALL);
	if (CURLE_OK != cc)
		diex("%s: fail: curl-global-init: %s", __func__,
				curl_easy_strerror(cc));

	assert(NULL == opts.cookie);
	if (NULL == (opts.cookie = curl_multi_init()))
		diex("%s: fail: multi-curl-init", __func__);

#ifdef CURLMOPT_PIPELININE
	cm = curl_multi_setopt(opts.cookie, CURLMOPT_PIPELINING, 1);
	if (CURLM_OK != cm)
		diex("%s: fail: multi-curl-setopt: %s", __func__,
				curl_multi_strerror(cm));
#else
	(void)cm;
#endif

	if (NULL == (opts.share = curl_share_init()))
		diex("%s: fail: share-curl-init", __func__);
	cs = curl_share_setopt(opts.share, CURLSHOPT_SHARE, 
			CURL_LOCK_DATA_DNS);
	if (CURLSHE_OK != cs)
		diex("%s: fail: share-curl-setopt", __func__);

	cs = curl_share_setopt(opts.share, CURLSHOPT_SHARE,
			CURL_LOCK_DATA_COOKIE);
	if (CURLSHE_OK != cs)
		diex("%s: fail: share-curl-setopt", __func__);
}


void
comm_teardown(void)
{
	CURLMcode 	mc;
	CURLSHcode 	ms;

	assert(NULL != opts.cookie);
	if (CURLM_OK != (mc = curl_multi_cleanup(opts.cookie))) 
		diex("%s: fail: multi-curl-cleanup: %s",
				__func__, curl_multi_strerror(mc));
	opts.cookie = NULL;

	assert(NULL != opts.share);
	if (CURLSHE_OK != (ms = curl_share_cleanup(opts.share)))
		diex("%s: fail: share-curl-cleanup: %s",
				__func__, curl_share_strerror(ms));
	opts.share = NULL;
	curl_global_cleanup();
}


void *
comm_report_next(int *run, int *type)
{
	CURLMsg		*msg;

	assert(NULL != opts.cookie);
	assert(NULL != run);
	assert(NULL != type);

	if (NULL == (msg = curl_multi_info_read(opts.cookie, run)))
		return(NULL);

	switch(msg->data.result) {
	case (CURLE_OK):
		*type = COMM_OK;
		break;
	case (CURLE_READ_ERROR):
	case (CURLE_WRITE_ERROR):
		*type = COMM_CALLBACK;
		break;
	case (CURLE_FTP_CANT_GET_HOST):
	case (CURLE_COULDNT_RESOLVE_PROXY):
	case (CURLE_COULDNT_RESOLVE_HOST):
		*type = COMM_RESOLVE;
		break;
	case (CURLE_FTP_CANT_RECONNECT):
	case (CURLE_COULDNT_CONNECT):
		*type = COMM_CONNECT;
		break;
	case (CURLE_OPERATION_TIMEDOUT):
		*type = COMM_TIMEOUT;
		break;
	case (CURLE_HTTP_RETURNED_ERROR):
		*type = COMM_SERVER;
		break;
	case (CURLE_URL_MALFORMAT):
		*type = COMM_SYNTAX;
		break;
	case (CURLE_OUT_OF_MEMORY):
	case (CURLE_ABORTED_BY_CALLBACK):
	case (CURLE_BAD_FUNCTION_ARGUMENT):
	case (CURLE_FILESIZE_EXCEEDED):
		*type = COMM_INTERNAL;
		break;
	default:
		*type = COMM_GENERAL;
		break;
	}
	return(msg->easy_handle);
}


int
comm_perform(void)
{
	CURLMcode	cm;
	int		run;

	assert(NULL != opts.cookie);
	for (;;)  {
		cm = curl_multi_perform(opts.cookie, &run);
		if (CURLM_CALL_MULTI_PERFORM != cm)
			break;
	}
	if (CURLM_OK == cm)
		return(run);

	diex("%s: fail: curl-multi-perform: %s", 
			__func__, curl_multi_strerror(cm));
	/* NOTREACHED */
}


int
comm_wait(void)
{
	fd_set		set;
	CURLMcode	cm;
	int		mfds;
	fd_set		r1, r2;
	struct timeval	tm;

	assert(NULL != opts.cookie);

	FD_ZERO(&set);
	FD_ZERO(&r1);
	FD_ZERO(&r2);

	cm = curl_multi_fdset(opts.cookie, &set, &r1, &r2, &mfds);
	if (CURLM_OK != cm)
		diex("%s: fail: multi-fdset: %s", __func__,
				curl_multi_strerror(cm));
	if (-1 == mfds)
		return(0);

	timerclear(&tm);
	tm.tv_sec = 0;
	tm.tv_usec = 100000;

	if (-1 == select(mfds, &set, NULL, NULL, &tm))
		die("%s: fail: select", __func__);
	return(1);
}


