/* $Id: comm.h,v 1.19 2007-05-29 16:05:57 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef COMM_H
#define COMM_H

__BEGIN_DECLS


struct comm_rep_t
{
	unsigned char	ip[NLPCR_IP_WIDTH];
	char		auth[NLPCR_URI_AUTHSZ];
	uint64_t 	expires;
	char 		charset[NLPCR_CHARSETLEN];
};


/*
 * Communication structure.  Note that this must fulfill functions for
 * both normal communications ("record" fetches) and robots.txt (REP)
 * fetches.  To this extent, the data union provides the central bits of
 * information as switched on the comm type.
 */
struct comm_t 
{
	char		uri[NLPCR_URI_MAXSZ]; 
	char		pretty[NLPCR_URI_FMTSZ]; 
	int		fd;
	uint32_t	status;
	void		*handle;
	void		*hash;
	
	union
	{
		struct nlpcdbrp	*pair;
		struct comm_rep_t rep;
	} data;

#define COMM_RECORD	(0)
#define COMM_REP	(1)
	uint8_t		type;

	struct 
	{
#define COMM_HEAD_SZ	(256)
		void	*handle;
		char	date[COMM_HEAD_SZ]; 
		char	agent[COMM_HEAD_SZ]; 
		char	lang[COMM_HEAD_SZ]; 
		char	charset[COMM_HEAD_SZ]; 
		char	host[COMM_HEAD_SZ + NLPCR_URI_AUTHSZ];
	} head;
};


struct comm_stats 
{
	uint64_t	ttime;		/* Total transfer time.      */
	uint64_t	nl_time;	/* Namelookup time.          */
	uint64_t	ct_time;	/* Connect time.             */
	uint64_t	pt_time;	/* Pre-transfer time.        */
	uint64_t	st_time;	/* Start-transfer time.      */
	uint64_t	dl_size;	/* Download size.            */
};


/*
 * Reset a comm handle between transfer sessions.  This may be used for
 * both records and REPs.
 */
#define comm_reset(p) do {                                           \
		void *sv = (p)->handle;                              \
		memset((p), 0, sizeof(struct comm_t));               \
		(p)->fd = -1;                                        \
		(p)->handle = sv;                                    \
	} while(0)


/*
 * Install a space-separated list of acceptable languages (this will be
 * terminated with the wild-card).  Note that the parameter val will be
 * changed by the callee.
 */
void
comm_hdr_lang(char *val);


/*
 * Install a space-separated list of acceptable charsets (this will be
 * terminated with the wild-card).  Note that the parameter val will be
 * changed by the callee.
 */
void
comm_hdr_charset(char *val);


/*
 * Install the user-agent.
 */
void
comm_hdr_agent(const char *val);


/*
 * Free up the memory held by a communicator as initialised in comm_new.
 * This may be used for both records and REPs.
 */
void
comm_free(struct comm_t *p);


/*
 * Allocate a communication object's internals.  Must be called before
 * any communication object functions.  Must be followed symmetrically
 * by a comm_free.  This may be used for both records and REPs.
 */
void
comm_new(struct comm_t *p);


/*
 * Close a transfer opened with comm_open.  Note that the object may
 * only be re-used after a call to comm_reset.  It may also only be used
 * for record transfers.  If the comm_stats structure is NULL, no
 * statistics will be returned (although they'll still be gathered).
 */
void
comm_close(struct comm_t *p, struct comm_stats *s);


/*
 * Same as comm_close, but for REPs.
 */
void
rep_close(struct comm_t *p, struct comm_stats *s);


/*
 * Open a transfer to an endpoint.  Prepares the local cache for writing
 * and all other miscellaneous activities involved in a transfer
 * transaction.  prefix is the cache directory prefix.  This may only be
 * used for record transfers.
 */
void
comm_open(struct comm_t *p, const char *prefix, struct nlpcdbrp *pair);


/*
 * Like comm_open but for REPs.
 */
void
rep_open(struct comm_t *p, const char *prefix, struct nlpcdbrp *pair);


/*
 * Initialise the communication subsystem.  This must be called before
 * any other functions are called.  The lang and enc dictionary pointers
 * may be NULL (see dict_has for details).  queue must be a valid
 * synchronised queue.  If set to ascend, new links will be path
 * ascended.
 */
void
comm_init(void *lang, void *enc, void *queue, int ascend);


/*
 * Teardown the communication subsystem.  May only be called after
 * comm_init.
 */
void
comm_teardown(void);


#define COMM_OK		(0)
#define COMM_CALLBACK	(1)
#define COMM_RESOLVE	(2)
#define COMM_CONNECT	(3)
#define COMM_TIMEOUT	(4)
#define COMM_SERVER	(5)
#define COMM_SYNTAX	(6)
#define COMM_INTERNAL	(7)
#define COMM_GENERAL	(8)

/*
 * Report on a completed or aborted transfer.  Sets "type" to be the
 * general status code of the transfer.  This is used for any time of
 * communication type.
 */
void *
comm_report_next(int *cookie, int *type);


/*
 * Perform work and scan for completed transfers.  Returns an
 * indentifier for the completed (or aborted) transfer.  This cookie can
 * be passed into comm_report_next in order to get further data.  This
 * is used for any time of communication type (non sequitur).
 */
int
comm_perform(void);


/*
 * Wait for data on any of the comm descriptors.  This internally uses
 * select.  FIXME - use a more efficient means of polling (like poll).
 * Returns 1 whether elements were waited upon, else 0.
 */
int
comm_wait(void);


__END_DECLS

#endif /*!COMM_H*/
