/* $Id: fetch.c,v 1.153 2007-05-30 14:51:33 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>

#include <assert.h>
#include <err.h>
#include <errno.h>
#include <getopt.h>
#include <inttypes.h>
#include <pthread.h>
#include <signal.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "nlpcrawl.h"
#include "comm.h"


/* 
 * Default number of active communication descriptors.  This refers to
 * the number of simultaneous transfers; the shell's ulimit must be able
 * to handle this value.
 */
#define FETCH_DEF_COMM	(512)

/*
 * Number of elements to hosts to allow to queue. It makes sense to have
 * this larger than the number of communication objects.
 */
#define FETCH_QUEUE_SZ	(1024)

/*
 * Number of connections (maximum) to simultaneously open per host.
 */
#ifdef CURLMOPT_PIPELINING
# define FETCH_QUEUE_BURST_SZ (6)
#else
# define FETCH_QUEUE_BURST_SZ (2)
#endif

/*
 * Maximum number of per-host fetch objects to queue.  This should be a
 * fairly large number: however, if too large, then a single entry can
 * dominate the connection time at (this * burst_wait) possible
 * overhead.
 */
#ifdef CURLMOPT_PIPELINING
# define FETCH_QUEUE_BURST_MAX (30)
#else
# define FETCH_QUEUE_BURST_MAX (10)
#endif

/*
 * Politeness policy wait delay (between bursts) as offset by the last
 * transfer sessions's end time.
 */
#define FETCH_QUEUE_BURST_WAIT (10)

/*
 * Age at which REP entities expire.  They'll be refetched on-demand at
 * this time (as per the draft).
 */ 
#define FETCH_REP_AGE		(60 * 60 * 24 * 7)

/*
 * Default maximum number of elements in the REP deny queue.  If set to
 * 0, the queue is unbounded.
 */
#define FETCH_REP_DENY	(1024)

/*
 * Default maximum number of elements in the REP allow queue.  If set to
 * 0, the queue is unbounded.
 */
#define FETCH_REP_ALLOW	(1024)


struct host_t 
{
	void		*queue;		/* Queue of perhost nlpcdbrp. */
	size_t		queue_sz;	/* Count in queue.            */
	time_t		last;		/* Last fetch from host.      */
	size_t		pending;	/* Number of pending fetches. */
	int		valid;		/* Whether referenced hostref */
					/* actively being used.       */
	unsigned char	ip[NLPCR_IP_WIDTH];
};


struct rep_t 
{
	time_t		age;
	time_t		expires;
	struct rep_t	*next;
	struct rep_t	*prev;
	void		*dallow;
	void		*ddeny;
	int		allow;
	int		deny;
	size_t		delay;
};


struct rcache_t
{
	void		*htab;
	struct rep_t	*first;
	struct rep_t	*last;
};


struct ccache_t
{
	void		*htab;
	void		*queue;
	struct comm_t	*array;
};


struct hcache_t
{
	void		*htab;
	void		*queue;
	struct host_t	*array;
};


static struct opt_t 
{
	size_t		ptrigger;	/* Fetch queue fill-trigger. */
	size_t		pburst;		/* Fetch burst size.         */
	size_t		pwidth;		/* Fetch queue width.        */
	size_t		pdepth;		/* Fetch queue depth.        */
	size_t		pwait;		/* Fetch wait seconds.       */
	size_t		ppar;		/* Fetch parallel comms.     */
	void		*wqueue;	/* Write queue.              */
	void		*rqueue;	/* Read queue.               */
	char		*cache;		/* Cache file directory.     */
	int		ascend;		/* Perform path-ascendence.  */
	void		*enc;		/* Dict of encodings.        */
	void		*lang;		/* Dict of languages.        */
	void		*auth;		/* Dict of authorities.      */
	void		*scheme;	/* Dict of schemes.          */
	void		*reps;		/* Hashtab of REP entities.  */
	char		*rcache;	/* Cached REP directory.     */
	size_t		rage;		/* Maximum age of REPs.      */
	size_t		maxallow;	/* Maximum allows per REP.   */
	size_t		maxdeny;	/* Maximum denies per REP.   */
} opts;


static void
rep_new(struct rep_t *r)
{
	assert(NULL != r);
	memset(r, 0, sizeof(struct rep_t));

	time_s(r->age);
	r->expires = r->age + (time_t)opts.rage;
	r->delay = opts.pwait;

	/* These are used for lookups. */
	dict_new(&r->ddeny);
	dict_new(&r->dallow);
}


static void
rep_free(struct rep_t *r)
{
	assert(NULL != r);
	assert(NULL != r->ddeny);
	assert(NULL != r->dallow);

	dict_free(&r->ddeny);
	dict_free(&r->dallow);
}


static int
rep_check(struct rep_t *r, struct uri *u)
{
	char		buf[NLPCR_URI_PATHSZ],
			addr[NLPCR_URI_FMTSZ];
	void		*seq;

	assert(NULL != r);
	assert(NULL != u);

	if (r->allow) {
		ddbg("%s: rep: all-allow: %s", __func__,
				uri_topstring_r(u, addr));
		return(1);
	}
	if (r->deny) {
		ddbg("%s: rep: all-deny: %s", __func__,
				uri_topstring_r(u, addr));
		return(0);
	}

	if (dict_has(&r->dallow, u->path, NLPCR_URI_PATHSZ)) {
		ddbg("%s: rep: path-exact-allow: %s", __func__,
				uri_topstring_r(u, addr));
		return(1);
	}
	if (dict_has(&r->ddeny, u->path, NLPCR_URI_PATHSZ)) {
		ddbg("%s: rep: path-exact-deny: %s", __func__,
				uri_topstring_r(u, addr));
		return(0);
	}

	dict_iter_begin(&r->dallow, &seq);
	while (dict_iter_next(&seq, buf)) {
		if (0 != strncmp(buf, u->path, strlen(buf)))
			continue;
		ddbg("%s: rep: path-part-allow: %s", __func__,
				uri_topstring_r(u, addr));
		dict_iter_close(&seq);
		return(1);
	}
	dict_iter_close(&seq);

	dict_iter_begin(&r->ddeny, &seq);
	while (dict_iter_next(&seq, buf)) {
		if (0 != strncmp(buf, u->path, strlen(buf)))
			continue;
		ddbg("%s: rep: path-part-deny: %s", __func__,
				uri_topstring_r(u, addr));
		dict_iter_close(&seq);
		return(0);
	}
	dict_iter_close(&seq);

	ddbg("%s: rep: default-allow: %s", __func__,
			uri_topstring_r(u, addr));

	return(1);
}


static int
rep_lookup(struct rcache_t *rcache, struct nlpcdbrp *pair,
		struct rep_t **r)
{
	time_t		cur;
	char		buf[NLPCR_URI_FMTSZ], *auth;

	assert(NULL != pair);
	assert(NULL != r);
	assert(NULL != rcache);
	assert(NULL != rcache->htab);

	auth = pair->record.uri.auth;
	if ( ! hash_get(&rcache->htab, auth,
				NLPCR_URI_AUTHSZ, (void **)r))
		return(0);

	assert(NULL != rcache->first);
	assert(NULL != rcache->last);

	time_s(cur);
	if (0 != (*r)->expires && (*r)->expires < cur) {
		ddbg("%s: robots.txt expired: %s", __func__,
				uri_topstring_r
				(&pair->record.uri, buf));
		hash_del(&opts.reps, auth, NLPCR_URI_AUTHSZ);

		if (NULL != (*r)->next)
			(*r)->next->prev = NULL == (*r)->prev ?
				NULL : (*r)->prev;
		if (NULL != (*r)->prev)
			(*r)->prev->next = NULL == (*r)->next ?
				NULL : (*r)->next;

		if (rcache->first == (*r))
			rcache->first = (*r)->next;
		if (rcache->last == (*r))
			rcache->last = (*r)->prev;
	
		rep_free(*r);
		free(*r);
		return(0);
	}

	return(1);
}


static void
thread_exec_fill(void *htab, void *queue, struct host_t *buf, 
		size_t *width, size_t *slots)
{
	struct nlpcdbrp	*p;
	struct host_t	*h;
	unsigned char	*ip;
	size_t		i;

	assert(NULL != queue);
	assert(NULL != htab);
	assert(NULL != buf);
	assert(NULL != width);
	assert(NULL != slots);

	assert(0 < opts.pdepth);

	for (p = NULL; ; ) {
		if (NULL != p)
			free(p);
		p = squeue_rem(opts.rqueue, SQUEUE_NOTIFY);
		if (NULL == p)
			break;
		ip = p->record.ip;

		if ( ! hash_get(&htab, ip, NLPCR_IP_WIDTH, 
					(void **)&h)) {
			if (*width >= opts.pwidth)
				continue;
			for (i = 0; i < opts.pwidth; i++)
				if ( ! buf[i].valid)
					break;
			assert(i < opts.pwidth);
			h = &buf[i];

			h->valid = 1;
			h->pending = 0;
			h->queue_sz = 0;
			h->last = 0;
			memcpy(h->ip, ip, NLPCR_IP_WIDTH);

			assert(NULL != h->queue);

			q_add(&queue, h);
			hash_put(&htab, ip, NLPCR_IP_WIDTH, h);
			++(*width);
		}
	
		if (h->queue_sz >= opts.pdepth)
			continue;

		q_add(&h->queue, p);
		++(h->queue_sz);
		++(*slots);
		p = NULL;
		if (*slots >= (opts.pdepth * opts.pwidth))
			break;
	}
}


static void
rep_postprocess(struct comm_t *p, struct rcache_t *rcache, int type)
{
	struct rep_t	*r;
	char		*buf;
	struct rep_parse_t parse;

	assert(NULL != p);
	assert(COMM_REP == p->type);

	/*
	 * If we had a rep being processed already, and it completed
	 * while our request was pending, simply allow the earlier
	 * version to be the true one.
	 */

	if (hash_get(&rcache->htab, p->data.rep.auth,
				NLPCR_URI_AUTHSZ, NULL)) {
		ddbg("%s: rep already acquired: %s", 
				__func__, p->pretty);
		rep_close(p, NULL);
		comm_reset(p);
		return;
	}

	calloc_s(r, 1, sizeof(struct rep_t));
	rep_new(r);
	rep_parse_init(opts.maxallow, opts.maxdeny, &parse);
	r->deny = 1;

	/*
	 * This follows the REP convention.  If there were errors during
	 * the transfer, deny all requests.  If there were non-200 and
	 * non-404 error codes, deny all requests.  If there was a 404
	 * request, allow all error codes.  Else, process the REP file's
	 * contents.
	 */

	switch (type) {
	case (COMM_OK):
		assert(200 == p->status);
		if (0 != p->data.rep.expires)
			r->expires = (time_t)p->data.rep.expires;
		if (rep_parse(p->pretty, p->fd, &parse))
			r->deny = r->allow = 0;
		break;
	case (COMM_CALLBACK):
		if (404 == p->status) {
			r->allow = 1;
			r->deny = 0;
		}
		break;
	case (COMM_INTERNAL):
		diex("%s: fail: internal: %s", __func__, p->pretty);
		/* NOTREACHED */
	default:
		break;
	}

	/* 
	 * TODO: this has double the memory it should.  dict_put copies
	 * the key and q_add is working on malloc'd content.  What
	 * really should exist is a dict_iter set of subroutines (ew).
	 */

	if ( ! r->deny && ! r->allow && (REP_ALLOW & parse.fields))
		while (NULL != (buf = q_rem(&parse.allow))) {
			dict_put(&r->dallow, buf, NLPCR_URI_PATHSZ);
			free(buf);
		}
	if ( ! r->deny && ! r->allow && (REP_DENY & parse.fields))
		while (NULL != (buf = q_rem(&parse.deny))) {
			dict_put(&r->ddeny, buf, NLPCR_URI_PATHSZ);
			free(buf);
		}
	if ( ! r->deny && ! r->allow && (REP_DELAY & parse.fields))
		r->delay = (size_t)parse.crawl_delay;

	if (NULL != rcache->first) {
		assert(NULL != rcache->last);
		r->next = rcache->first;
		rcache->first->prev = r;
		rcache->first = r;
	} else {
		assert(NULL == rcache->last);
		rcache->first = rcache->last = r;
	}

	hash_put(&rcache->htab, p->data.rep.auth, NLPCR_URI_AUTHSZ, r);

	rep_parse_free(&parse);
	rep_close(p, NULL);
}


static void
comm_postprocess(struct comm_t *p, int type)
{
	uint64_t	tm;
	struct nlpcdbr	*r;

	assert(NULL != p);
	assert(COMM_RECORD == p->type);
	assert(NULL != p->data.pair);

	/*
	 * Simple routine to service a nlpcdbrp record as demanded by
	 * the transfer status codes.  Also updates the timestamps on
	 * the relevant record.  When finished, puts the record into the
	 * processing queue to be serviced in the write thread.
	 */

	tm = time_uint64();
	r = &p->data.pair->record;

	r->atime = r->mtime = tm;
	if (COMM_OK == type)
		r->ftime = tm;

#define set_state(st, cd) r->state = (st); r->code = (cd)

	switch (type) {
	case (COMM_OK):
		set_state(NLPCDBR_FRESH, NLPCDBR_OK);
		break;
	case (COMM_CALLBACK):
		break;
	case (COMM_RESOLVE):
		set_state(NLPCDBR_UNREACH, NLPCDBR_RESOLV);
		break;
	case (COMM_CONNECT):
		set_state(NLPCDBR_UNREACH, NLPCDBR_CONNECT);
		break;
	case (COMM_TIMEOUT):
		set_state(NLPCDBR_UNREACH, NLPCDBR_TIMEO);
		break;
	case (COMM_SERVER):
		set_state(NLPCDBR_UNREACH, NLPCDBR_SERVER);
		break;
	case (COMM_SYNTAX):
		set_state(NLPCDBR_DEAD, NLPCDBR_URL);
		break;
	case (COMM_INTERNAL):
		diex("%s: fail: internal: %s", __func__, p->pretty);
		/* NOTREACHED */
	default:
		set_state(NLPCDBR_UNREACH, NLPCDBR_GENERAL);
		break;
	}

	comm_close(p, NULL);
	squeue_add(opts.wqueue, p->data.pair, 
			SQUEUE_NOTIFY | SQUEUE_WAIT);
}


/*
 * Cleanup routine for the comm cache.
 */
static void
thread_exec_cleanup_ccache(void *p)
{
	size_t		i;
	struct ccache_t	*ccache;

	assert(NULL != p);
	ccache = (struct ccache_t *)p;

	for (i = 0; i < opts.ppar; i++)
		comm_free(&ccache->array[i]);

	free(ccache->array);
	hash_free(&ccache->htab);
	q_free(&ccache->queue);
}


/*
 * Cleanup routine for the host cache.
 */
static void
thread_exec_cleanup_hcache(void *p)
{
	size_t		i;
	struct hcache_t	*hcache;

	assert(NULL != p);
	hcache = (struct hcache_t *)p;

	for (i = 0; i < opts.pwidth; i++)
		q_free(&hcache->array[i].queue);

	free(hcache->array);
	hash_free(&hcache->htab);
	q_free(&hcache->queue);
}


/*
 * Cleanup routine for the REP cache.
 */
static void
thread_exec_cleanup_rcache(void *p)
{
	struct rcache_t	*rcache;
	struct rep_t	*r, *r2;

	assert(NULL != p);
	rcache = (struct rcache_t *)p;
	for (r = rcache->first; NULL != r; ) {
		r2 = r;
		r = r->next;
		rep_free(r2);
		free(r2);
	}

	hash_free(&rcache->htab);
}


void *
thread_exec(void *arg)
{
	size_t		i, slots, width, pend;
	void		*handle;
	struct host_t	*h;
	struct comm_t	*c;
	struct hcache_t	hcache;
	struct rcache_t	rcache;
	struct ccache_t	ccache;
	struct rep_t	*rep;
	struct nlpcdbrp	*p;
	time_t		cur;
	int		type, cookie;
	unsigned char	ip[NLPCR_IP_WIDTH];
	char		addr[NLPCR_URI_FMTSZ];

	assert(NULL == arg);

	/*
	 * In order to keep things efficient (the main loop shouldn't be
	 * operating with subcomponents of over constant- or at most
	 * logarithmic-time overhead), we need a variety of structures.
	 * ccache and hcache have an array-backed queue and hashtable:
	 * "waiting" elements are in the queue, while the hashtable may
	 * be queried in constant time.  The backing array may be
	 * cleaned up later (the underlying memory pool is fixed).
	 * rcache is backed by its linked list.  It has a hashtable
	 * referencing the nodes for fast access.
	 */

	comm_init(opts.lang, opts.enc, opts.wqueue, opts.ascend);

	calloc_s(ccache.array, opts.ppar, sizeof(struct comm_t));
	calloc_s(hcache.array, opts.pwidth, sizeof(struct host_t));

	rcache.first = rcache.last = NULL;

	q_new(&hcache.queue);
	q_new(&ccache.queue);

	hash_new(&rcache.htab);
	hash_new(&hcache.htab);
	hash_new(&ccache.htab);

	for (i = 0; i < opts.ppar; i++) {
		comm_new(&ccache.array[i]);
		q_add(&ccache.queue, &ccache.array[i]);
		hash_put(&ccache.htab, &ccache.array[i].handle, 
				sizeof(void *), &ccache.array[i]);
	}

	for (i = 0; i < opts.pwidth; i++)
		q_new(&hcache.array[i].queue);

	/* 
	 * All memory should be reclaimed at the end of the loop.
	 */

	pthread_cleanup_push((void(*)(void *))comm_teardown, 
			(void *)NULL);
	pthread_cleanup_push(thread_exec_cleanup_ccache, 
			(void *)&ccache);
	pthread_cleanup_push(thread_exec_cleanup_hcache, 
			(void *)&hcache);
	pthread_cleanup_push(thread_exec_cleanup_rcache, 
			(void *)&rcache);

	slots = 0;
	pend = 0;
	width = 0;

	/*
	 * We must have at least one physical host spot before
	 * continuing with the algorithm.
	 */

	while (width == 0) {
		sleep(1);
		thread_exec_fill(hcache.htab, hcache.queue, 
				hcache.array, &width, &slots);
	}

thread_exec_head:

	/*
	 * Fill the physical host cache.  Only enter the fill routines
	 * if we're below a threshold.  If we are, it follows that we
	 * may have zero entries; if we do, wait until we have at least
	 * one before continuing with the algorithm.
	 */

	if (width < opts.ptrigger) for (;;) {
		thread_exec_fill(hcache.htab, hcache.queue, 
				hcache.array, &width, &slots);
		if (0 < slots)
			break;
		sleep(1);
	}

	time_s(cur);

	/* 
	 * Scheduling loop.  Pops all available connections from the
	 * connection queue and tries to assign targets to them.  This
	 * must take into consideration REP and other measures.
	 */

	for ( ; pend < opts.ppar; ) {
		h = (struct host_t *)q_head(&hcache.queue);
		if (NULL == h)
			break;

		/*
		 * Raw politeness policy:  each physical host has an
		 * associated politeness as dictated by command-line
		 * options (or defaults).  Check it here.
		 */

		if (0 != h->last && cur < h->last + (time_t)opts.pwait)
			break;
		h = q_rem(&hcache.queue);
		assert(NULL != h);

		/*
		 * "Burst" connections to the physical host by iterating
		 * over (possibly-) virtual hosts addressed at the
		 * physical host.  First, cross-check the (possibly-)
		 * REP assigned to the host (remember that REPs are
		 * assigned to virtual hosts, not physical ones).
		 */

		while (opts.pburst > h->pending && 0 < h->queue_sz) {
			p = q_head(&h->queue);
			if ( ! rep_lookup(&rcache, p, &rep)) {
				/*
				 * No REP found for the host.  Schedule
				 * one to the connection ostensively
				 * allocated for a record fetch.
				 */
				c = q_rem(&ccache.queue);
				assert(NULL != c);
				++(h->pending);
				rep_open(c, opts.rcache, p);
				dbg("scheduling rep (%zd, %zd, %zd): "
						"%s", pend, slots, 
						width, c->pretty);
				++pend;
				break;
			}

			if ( ! rep_check(rep, &p->record.uri)) {
				q_rem(&h->queue);
				--(h->queue_sz);
				p->record.code = NLPCDBR_REP;
				p->record.state = NLPCDBR_UNREACH;
				squeue_add(opts.wqueue, p,
					SQUEUE_NOTIFY | SQUEUE_WAIT);

				ddbg("%s: rep violation: %s", __func__,
						uri_topstring_r
						(&p->record.uri, addr));
				continue;
			}

			if (cur < h->last + (time_t)rep->delay &&
					0 != h->last)
				break;

			q_rem(&h->queue);

			assert(NULL != p);
			--(h->queue_sz);
			++(h->pending);

			p->record.size = 0;
			p->record.expires = 0;
			p->record.type = NLPCDBR_TYPE_NONE;
			memset(p->record.charset, 0, NLPCR_CHARSETLEN);
			memset(p->record.lang, 0, NLPCR_LANGLEN);

			c = q_rem(&ccache.queue);
			assert(NULL != c);
			comm_open(c, opts.cache, p);

			dbg("scheduling record (%zd, %zd, %zd): %s", 
					pend, slots, width, c->pretty);
			if (opts.ppar <= ++pend)
				break;
		}
	}

	if ( ! comm_wait() && 0 == pend)
		sleep(1);

	pthread_testcancel();

	cookie = comm_perform();
	time_s(cur);

	while (NULL != (handle = comm_report_next(&cookie, &type))) {
		hash_get(&ccache.htab, &handle, sizeof(void *), 
				(void **)&c);
		assert(NULL != c);

		if (COMM_REP == c->type) {
			dbg("unscheduling rep (%zd, %zd, %zd): %s", 
					pend, slots, width, c->pretty);
			memcpy(ip, c->data.rep.ip, NLPCR_IP_WIDTH);
			rep_postprocess(c, &rcache, type);
		} else {
			assert(COMM_RECORD == c->type);
			assert(NULL != c->data.pair);
			dbg("unscheduling record (%zd, %zd, %zd): %s", 
					pend, slots, width, c->pretty);
			memcpy(ip, c->data.pair->record.ip, 
					NLPCR_IP_WIDTH);
			--slots;
			comm_postprocess(c, type);
		}
		comm_reset(c);
		--pend;

		hash_get(&hcache.htab, ip, NLPCR_IP_WIDTH, 
				(void **)&h);
		assert(NULL != h);
		--(h->pending);
		h->last = cur;

		if (0 == h->pending && 0 == h->queue_sz) {
			--width;
			h->valid = 0;
			hash_del(&hcache.htab, ip, NLPCR_IP_WIDTH);
		} else
			q_add(&hcache.queue, h);

		q_add(&ccache.queue, c);
	}

	goto thread_exec_head;

	/* NOTREACHED */
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);

	return(NULL);
}


void *
thread_write(void *arg)
{
	struct nlpcdbrp	*p;
	struct uri	*addr;
	char		buf[NLPCR_URI_FMTSZ],
			reason[32];
	size_t		n, u, d, f;
	uint64_t	wflags, rflags;

	assert(NULL == arg);

	wflags = SQUEUE_WAIT | SQUEUE_NOTIFY;
	rflags = SQUEUE_NOTIFY;
	u = f = n = d = 0;

	for (;;) {
		p = squeue_rem(opts.wqueue, rflags);
		if (NULL == p) {
			if (u > 0) 
				signal_next(NLPCDBR_UNREACH);
			if (f > 0)
				signal_next(NLPCDBR_FRESH);
			if (n > 0)
				signal_next(NLPCDBR_NEW);
			if (d > 0)
				signal_next(NLPCDBR_DEAD);
			u = d = f = n = 0;
			p = squeue_rem(opts.wqueue, wflags);
		}

		assert(NULL != p);
		addr = &p->record.uri;
		uri_topstring_r(addr, buf);
		nlpcdbr_code2str_r(reason, p->record.code);

		if (NLPCDBR_NEW == p->record.state &&
				0 == addr->scheme[0]) {
			dbg("reject: no scheme: %s", buf);
			goto thread_write_tail;
		}
		if (NLPCDBR_NEW == p->record.state &&
				0 == addr->auth[0]) {
			dbg("reject: no authority: %s", buf);
			goto thread_write_tail;
		}

		if (NLPCDBR_NEW == p->record.state && ! dict_has
				(&opts.scheme, addr->scheme, 
				 strlen(addr->scheme))) {
			dbg("reject: non-matching scheme: %s", buf);
			goto thread_write_tail;
		}
		if (NLPCDBR_NEW == p->record.state &&  ! dict_has
				(&opts.auth, addr->auth,
				 strlen(addr->auth))) {
			dbg("reject: non-matching auth: %s", buf);
			goto thread_write_tail;
		}

		if (NLPCDBR_NEW == p->record.state) {
			if ( ! db_add(p)) {
				ddbg("reject: duplicate: %s", buf);
				goto thread_write_tail;
			}
		} else if ( ! db_put(p))
			diex("%s: fail: db_put: %s", __func__, buf);

		switch (p->record.state) {
		case (NLPCDBR_UNREACH):
			u++;
			dbg("accept: unreach (%s): %s", reason, buf);
			break;
		case (NLPCDBR_FRESH):
			f++;
			dbg("accept: fresh: %s", buf);
			break;
		case (NLPCDBR_NEW):
			n++;
			dbg("accept: new: %s", buf);
			break;
		case (NLPCDBR_DEAD):
			d++;
			dbg("accept: dead (%s): %s", reason, buf);
			break;
		default:
			abort();
		}

thread_write_tail:	
		free(p);
	}

	/* NOTREACHED */
	return(NULL);
}


static int
scan(void)
{
	struct nlpcdbrp	*p;
	struct db_query	query;
	void		*cookie;
	int		flags, rc;
	static uint64_t	last;

	memset(&query, 0, sizeof(struct db_query));

	query.field = DB_FIELD_STATE;
	query.state = NLPCDBR_READY;
	query.skip = last;

	db_window_fill(&query);
	db_query_begin(&cookie);

	for (rc = 1, p = NULL, flags = 0; ; ) {
		if (SIGNAL_EXIT & (flags |= signal_test())) 
			break;
		calloc_s(p, 1, sizeof(struct nlpcdbrp));
		if (1 != (rc = db_query_next(&cookie, &query, p)))
			break;
		if ( ! squeue_add(opts.rqueue, p, SQUEUE_NOTIFY))
			break;
		p = NULL;
	}

	db_query_close(&cookie);
	last = (0 == rc ? 0 : query.pos);
	if (NULL != p)
		free(p);

	return(flags);
}


int
main(int argc, char *argv[])
{
	int 		c, t;
	char 		*dir;

	extern char	*__progname, *optarg;
	extern int	optind;
	pthread_t	writer, execer;

	static struct option longopts[] = {
		{ "head-lang", 		required_argument, NULL, 1 },
		{ "head-charset", 	required_argument, NULL, 2 },
		{ "filter-charset", 	required_argument, NULL, 3 },
		{ "filter-lang", 	required_argument, NULL, 4 },
		{ "filter-scheme", 	required_argument, NULL, 5 },
		{ "filter-auth", 	required_argument, NULL, 6 },
		{ "filter-ascend", 	no_argument, NULL, 7 },
		{ "fetch-parallel",	required_argument, NULL, 8 },
		{ "fetch-burst",	required_argument, NULL, 9 },
		{ "fetch-depth",	required_argument, NULL, 10 },
		{ "fetch-width",	required_argument, NULL, 11 },
		{ "fetch-wait",		required_argument, NULL, 12 },
		{ "fetch-parallel",	required_argument, NULL, 13 },
		{ "rep-expires",	required_argument, NULL, 14 },
		{ "rep-maxallow",	required_argument, NULL, 15 },
		{ "rep-maxdeny",	required_argument, NULL, 16 },
		{ NULL, 0, NULL, 0 }
	};

	memset(&opts, 0, sizeof(struct opt_t));

	opts.cache = NLPC_DEFAULT_FILES;
	opts.pburst = FETCH_QUEUE_BURST_SZ;
	opts.pdepth = FETCH_QUEUE_BURST_MAX;
	opts.pwait = FETCH_QUEUE_BURST_WAIT;
	opts.pwidth = FETCH_QUEUE_SZ;
	opts.ppar = FETCH_DEF_COMM;
	opts.rcache = NLPC_DEFAULT_REPS;
	opts.rage = FETCH_REP_AGE;
	opts.maxallow = FETCH_REP_ALLOW;
	opts.maxdeny = FETCH_REP_DENY;

	dir = NLPC_DEFAULT_DIR;

	while (-1 != (c = getopt_long(argc, argv, "hdvf:p:r:", 
					longopts, NULL))) 
		switch (c) {
		case (1):
			comm_hdr_lang(optarg);
			break;
		case (2):
			comm_hdr_charset(optarg);
			break;
		case (3):
			dict_free(&opts.enc);
			dict_fill(&opts.enc, optarg);
			break;
		case (4):
			dict_free(&opts.lang);
			dict_fill(&opts.lang, optarg);
			break;
		case (5):
			dict_free(&opts.scheme);
			dict_fill(&opts.scheme, optarg);
			break;
		case (6):
			dict_free(&opts.auth);
			dict_fill(&opts.auth, optarg);
			break;
		case (7):
			opts.ascend = 1;
			break;
		case (8):
			if ( ! parse_int(optarg, &t, INT_MAX, 1))
				diex("invalid value: --fetch-trigger");
			opts.ptrigger = (size_t)t;
			break;
		case (9):
			if ( ! parse_int(optarg, &t, INT_MAX, 1))
				diex("invalid value: --fetch-burst");
			opts.pburst = (size_t)t;
			break;
		case (10):
			if ( ! parse_int(optarg, &t, INT_MAX, 1))
				diex("invalid value: --fetch-depth");
			opts.pdepth = (size_t)t;
			break;
		case (11):
			if ( ! parse_int(optarg, &t, INT_MAX, 1))
				diex("invalid value: --fetch-width");
			opts.pwidth = (size_t)t;
			break;
		case (12):
			if ( ! parse_int(optarg, &t, INT_MAX, 0))
				diex("invalid value: --fetch-wait");
			opts.pwait = (size_t)t;
			break;
		case (13):
			if ( ! parse_int(optarg, &t, INT_MAX, 0))
				diex("invalid value: --fetch-parallel");
			opts.pwait = (size_t)t;
			break;
		case (14):
			if ( ! parse_int(optarg, &t, INT_MAX, 0))
				diex("invalid value: --rep-expires");
			opts.rage = (size_t)t;
			break;
		case (15):
			if ( ! parse_int(optarg, &t, INT_MAX, 0))
				diex("invalid value: --rep-maxallow");
			opts.maxallow = (size_t)t;
			break;
		case (16):
			if ( ! parse_int(optarg, &t, INT_MAX, 0))
				diex("invalid value: --rep-maxdeny");
			opts.maxdeny = (size_t)t;
			break;
		case ('d'):
			dbg_lvl++;
			break;
		case ('f'):
			opts.cache = optarg;
			break;
		case ('p'):
			dir = optarg;
			break;
		case ('r'):
			opts.rcache = optarg;
			break;
		case ('v'):
			fprintf(stdout, "%s version %s\n", 
					__progname, VERSION);
			return(EXIT_SUCCESS);
		case ('h'):
			/*FALLTHROUGH*/
		default:
			fprintf(stderr, "usage: %s [-dhv] [-f path] "
					"[-p path] [LONG_OPTS]\n", 
					__progname);
			return(EXIT_FAILURE);
		}
	argc -= optind;
	argv += optind;

	assert(opts.pwidth > 0);
	assert(opts.pdepth > 0);
	assert(opts.ppar > 0);
	assert(opts.pburst > 0);
	assert(opts.pwait > 0);

	if (0 == opts.ptrigger)
		opts.ptrigger = opts.pwidth / 4;
	if (opts.ptrigger > (opts.pwidth * opts.pdepth))
		diex("value overruns queue domain: --fetch-trigger");

	opts.wqueue = squeue_new(opts.pdepth * opts.pwidth);
	opts.rqueue = squeue_new(opts.pdepth * opts.pwidth);

	signal_init(dir, NLPCDBR_READY);
	db_init(dir, 0);

	if (0 != pthread_create(&writer, NULL, thread_write, NULL))
		die("%s: fail: pthread_create", __func__);
	if (0 != pthread_create(&execer, NULL, thread_exec, NULL))
		die("%s: fail: pthread_create", __func__);

	comm_hdr_agent(VERSION);

	for (;;) {
		if (SIGNAL_EXIT & scan())
			break;
		if (SIGNAL_EXIT & signal_wait(0))
			break;
	}

	if (0 != pthread_cancel(writer))
		die("%s: fail: pthread_cancel", __func__);
	if (0 != pthread_cancel(execer))
		die("%s: fail: pthread_cancel", __func__);

	if (0 != pthread_join(writer, NULL))
		die("%s: fail: pthread_join", __func__);
	if (0 != pthread_join(execer, NULL))
		die("%s: fail: pthread_join", __func__);

	db_teardown();
	signal_teardown();

	dict_free(&opts.enc);
	dict_free(&opts.lang);
	dict_free(&opts.auth);
	dict_free(&opts.scheme);

	squeue_free(opts.wqueue);
	squeue_free(opts.rqueue);

	return(EXIT_SUCCESS);
}

