/* $Id: hash.c,v 1.22 2007-05-30 14:51:33 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>

#include <assert.h>
#include <err.h>
#include <ctype.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <db.h>

#include "nlpcrawl.h"


static int
_del(void **arg, void *key, size_t ksz)
{
	DBT		dbk;

	assert(NULL != arg);
	assert(NULL != key);

	if (0 == ksz)
		return(DB_NOTFOUND);

	memset(&dbk, 0, sizeof(DBT));

	dbk.data = key;
	dbk.size = ksz;

	if (NULL == *arg)
		return(0);
	return((*(DB **)arg)->del(*(DB **)arg, NULL, &dbk, 0));
}


static int
_get(void **arg, void *key, size_t ksz, void **val)
{
	DBT		dbk, dbv;
	int		rc;

	assert(NULL != arg);
	assert(NULL != key);

	if (0 == ksz)
		return(DB_NOTFOUND);

	memset(&dbk, 0, sizeof(DBT));
	memset(&dbv, 0, sizeof(DBT));

	dbk.data = key;
	dbk.size = ksz;

	if (NULL == *arg)
		return(0);

	rc = (*(DB **)arg)->get(*(DB **)arg, NULL, &dbk, &dbv, 0);
	if (0 == rc && NULL != val)
		*val = *(void **)dbv.data;
	return(rc);
}


static int
_put(DB *db, void *key, size_t sz, void *val)
{
	DBT		dbk, dbv;

	assert(NULL != db);
	assert(NULL != key);
	assert(0 < sz);

	memset(&dbk, 0, sizeof(DBT));
	memset(&dbv, 0, sizeof(DBT));

	dbk.data = key;
	dbk.size = sz;

	if (NULL != val) {
		dbv.data = &val;
		dbv.size = sizeof(val);
	}

	return(db->put(db, NULL, &dbk, &dbv, 0));
}


int
hash_del(void **arg, void *key, size_t ksz)
{
	int		rc;

	if (0 == (rc = _del(arg, key, ksz)))
		return(1);
	else if (DB_NOTFOUND == rc)
		return(0);

	diex("%s: fail: hashtable del: %s", 
			__func__, db_strerror(rc));
	/* NOTREACHED */
}


int
hash_get(void **arg, void *key, size_t ksz, void **val)
{
	int		rc;

	if (NULL != val)
		*val = NULL;

	if (0 == (rc = _get(arg, key, ksz, val)))
		return(1);
	else if (DB_NOTFOUND == rc)
		return(0);

	diex("fail: hashtable get: %s", db_strerror(rc));
	/*NOTREACHED*/
}


int
dict_has(void **arg, void *key, size_t ksz)
{
	int		rc;

	assert(NULL != arg);
	assert(NULL != key);

	if (NULL == *arg)
		return(1);
	
	if (0 == (rc = _get(arg, key, ksz, NULL)))
		return(1);
	else if (DB_NOTFOUND == rc)
		return(0);

	diex("fail: dictionary has: %s", db_strerror(rc));
	/*NOTREACHED*/
}


int
dict_iter_next(void **seq, void *res)
{
	DBC		*dbc;
	DBT		key, val;
	int		rc;

	assert(NULL != seq);
	assert(NULL != *seq);
	assert(NULL != res);

	dbc = *(DBC **)seq;

	memset(&key, 0, sizeof(DBT));
	memset(&val, 0, sizeof(DBT));

	rc = dbc->c_get(dbc, &key, &val, DB_NEXT);
	memcpy(res, key.data, key.size);
	if (0 == rc)
		return(1);
	if (DB_NOTFOUND == rc)
		return(0);

	diex("%s: fail: cursor next: %s", 
			__func__, db_strerror(rc));
	/* NOTREACHED */
}


void
dict_iter_begin(void **cookie, void **seq)
{
	DBC		*dbc;
	int		rc;

	assert(NULL != cookie);
	assert(NULL != *cookie);
	assert(NULL != seq);

	rc = (*(DB **)cookie)->cursor(*(DB **)cookie, NULL, &dbc, 0);
	*seq = dbc;

	if (0 == rc)
		return;

	diex("%s: fail: cursor open: %s", 
			__func__, db_strerror(rc));
	/* NOTREACHED */
}


void
dict_iter_close(void **seq)
{
	DBC		*dbc;
	int		rc;

	assert(NULL != seq);
	assert(NULL != *seq);

	dbc = *(DBC **)seq;

	if (0 == (rc = dbc->c_close(dbc)))
		return;

	diex("%s: fail: cursor close: %s", 
			__func__, db_strerror(rc));
	/* NOTREACHED */
}


uint64_t
dict_keys(void **cookie)
{
	DBC		*dbc;
	DBT		key, val;
	uint64_t	c;
	int		rc;

	assert(NULL != cookie);

	rc = (*(DB **)cookie)->cursor(*(DB **)cookie, NULL, &dbc, 0);
	if (0 != rc)
		diex("%s: fail: dictionary cursor: %s", 
				__func__, db_strerror(rc));

	c = 0;
	for (c = 0; 0 != dbc->c_get(dbc, &key, &val, DB_NEXT); c++)
		/* Do nothing. */ ;
	dbc->c_close(dbc);
	return(c);
}


int
dict_file(void **cookie, const char *fname)
{
	char		line[HASH_FILE_WORD], c;
	FILE 		*in;
	int		rc;
	size_t		i;

	assert(NULL != cookie);
	assert(NULL != fname);

	if (NULL == (in = fopen(fname, "rb"))) {
		dddbg("%s: fail: open: %s", __func__, fname);
		return(0);
	}

	c = 1;
	dict_new(cookie);
	for (i = 1; i < HASH_FILE_WORD - 1 && EOF != c; ) {
		c = fgetc(in);
		/* Newline is terminator. */
		line[i] = (EOF == c || '\n' == c) ? '\0' : c;
		if ('\0' != line[i]) {
			i++;
			continue;
		} else if (i == 0)
			continue;

		if (0 != (rc = _put(*(DB **)cookie, line, i, NULL))) 
			if (DB_KEYEXIST != rc) {
				dddbg("%s: fail put: %s", __func__,
						db_strerror(rc));
				(void)fclose(in);
				return(0);
			}
		i = 0;
	}

	if (-1 == fclose(in)) {
		dddbg("%s: fail: close: %s", __func__, fname);
		return(0);
	}
	if (HASH_FILE_WORD - 1 == i) {
		dddbg("%s: fail: line too long: %s", __func__, fname);
		return(0);
	}

	return(1);
}


void
dict_fill(void **cookie, char *str)
{
	char 		*t, *tt;

	assert(NULL != cookie);
	assert(NULL != str);

	dict_new(cookie);

	if (NULL == *cookie)
		diex("fail: hash cookie is un-set");

	while (NULL != (t = strsep(&str, " "))) { 
		if (0 == strlen(t))
			continue;
		for (tt = t; 0 != *tt; tt++) 
			if (isalpha(*tt))
				*tt = (char)tolower(*tt);
		dddbg("%s: %s", __func__, t);
		if (0 != _put(*(DB **)cookie, t, strlen(t), NULL))
			diex("fail: hash put");
	} 
} 


void
hash_put(void **arg, void *key, size_t ksz, void *val)
{
	int		rc;

	assert(NULL != arg);
	assert(NULL != key);
	assert(NULL != val);
	assert(0 < ksz);
	assert(NULL != *arg);

	if (0 == (rc = _put(*(DB **)arg, key, ksz, val)))
		return;
	diex("fail: hashtable put: %s", db_strerror(rc));

	/*NOTREACHED*/
}


void
dict_put(void **arg, void *key, size_t sz)
{
	int		rc;

	assert(NULL != arg);
	assert(NULL != key);
	assert(0 < sz);
	assert(NULL != *arg);

	if (0 == (rc = _put(*(DB **)arg, key, sz, NULL)))
		return;
	diex("fail: dictionary put: %s", db_strerror(rc));

	/*NOTREACHED*/
}


void
dict_free(void **arg)
{
	int		rc;

	assert(NULL != arg);

	if (NULL == *arg) 
		return;

	rc = (*(DB **)arg)->close(*(DB **)arg, 0);
	*arg = NULL;

	if (0 == rc)
		return;
	diex("fail: dictionary close: %s", db_strerror(rc));
	/*NOTREACHED*/
}


void
hash_free(void **arg)
{
	dict_free(arg);
}


void
hash_new(void **cookie)
{
	dict_new(cookie);
}


void
dict_new(void **cookie)
{
	int		rc;

	assert(NULL != cookie);

	*cookie = NULL;
	rc = db_create((DB **)cookie, NULL, 0);

	if (0 != rc)
		diex("fail: dictionary create: %s", db_strerror(rc));

	rc = (*(DB **)cookie)->open(*(DB **)cookie, NULL, 
			NULL, NULL, DB_HASH, DB_CREATE, 0);

	if (0 == rc)
		return;
	diex("fail: dictionary open: %s", db_strerror(rc));
	/*NOTREACHED*/
}

