/* $Id: http.c,v 1.7 2007-05-29 16:05:57 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>

#include <assert.h>
#include <ctype.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "nlpcrawl.h"

static int
http_parse_pragma(struct http_header *head)
{
	assert(NULL != head);
	assert(NULL != head->val);

	head->data.pragma.pragma_type = HTTP_PRAGMA_UNKNOWN;
	if (0 == strcasecmp("no-cache", head->val)) {
		head->data.pragma.pragma_type = HTTP_PRAGMA_NOCACHE;
		head->data.pragma.data.nocache = 1;
	}

	return(1);
}

static int
http_parse_content_type(struct http_header *head)
{
	size_t		sz, valsz, pos;
	char		*p;

	assert(NULL != head);
	assert(NULL != head->val);

	head->data.ctype.media = NLPCDBR_TYPE_NONE;
	head->data.ctype.charset = NULL;

#define HEADER_CONTENT_TYPE(str, res) \
	do if ((sz = strlen((str))) <= strlen(head->val) && \
			0 == strncasecmp(head->val, (str), sz)) { \
		head->data.ctype.media = (res); \
		goto parse_charset; \
	} while(0)

	if (';' == head->val[0])
		return(0);

	HEADER_CONTENT_TYPE("text/html", 
			NLPCDBR_TYPE_TXT_HTML);
	HEADER_CONTENT_TYPE("text/xml", 
			NLPCDBR_TYPE_TXT_XML);
	HEADER_CONTENT_TYPE("application/xhtml+html", 
			NLPCDBR_TYPE_APP_XHTML);
	HEADER_CONTENT_TYPE("text/plain", 
			NLPCDBR_TYPE_TXT_PLAIN);
	HEADER_CONTENT_TYPE("application/xml", 
			NLPCDBR_TYPE_APP_XML);

parse_charset:

	valsz = strlen(head->val);

	/* Skip ahead to semicolon. */
	if (NULL == (p = memchr(head->val, ';', valsz)))
		return(1);

	/* Nullify the semicolon. */
	pos = (p - head->val);

	*p++ = 0;
	pos++;

	/* Skip LWS. */
	for ( ; pos < valsz; p++, pos++)
		if (13 != *p && 10 != *p && 32 != *p && 9 != *p)
			break;

	/* Skip "charset=" token. */
	if ((sz = strlen("charset=")) > valsz - pos)
		return(0);
	if (0 != strncasecmp(p, "charset=", sz)) 
		return(0);
	pos += sz;
	p += sz;

	head->data.ctype.charset = p;
	if (0 == strlen(p))
		return(0);

	for ( ; 0 != *p; p++)
		if (isalpha(*p))
			*p = (char)tolower(*p);

	return(1);
}

int
http_parse(char *p, size_t sz, struct http_header *head)
{
	size_t		i, keysz, valsz;

	assert(NULL != head);
	assert(NULL != p);

	memset(head, 0, sizeof(struct http_header));

	if (0 == sz) 
		return(0);
	if ( ! isascii(*p))
		return(0);

	/* Read ahead to colon. */
	head->key = p;
	for (keysz = 0; ':' != *p && keysz < sz; p++, keysz++)  
		if (isascii(*p))
			*p = (char)tolower(*p);
	if (':' != *p)
		return(0);
	/* Nil-terminate the key. */
	*p++ = 0;

	if (keysz == sz)
		return(0);

	/* Skip LWS. */
	for (i = keysz + 1; i < sz; p++, i++) 
		if (13 != *p && 10 != *p && 32 != *p && 9 != *p)
			break;
	if (i == sz)
		return(0);

	head->val = p;
	valsz = (sz - i);

	p = head->val + valsz - 1;
	/* Nil-ify trailing LWS. */
	for (i = valsz - 1; (ssize_t)i >= 0; i--, p--)
		if (13 == *p || 10 == *p || 32 == *p || 9 == *p) {
			*p = 0;
			valsz--;
		} else 
			break;
	if (0 != head->val[valsz]) 
		return(0);

	if (0 == strcmp("expires", head->key)) {
		if (0 == strlen(head->val))
			return(0);
		head->type = HTTP_EXPIRES;
		head->data.expires = head->val;
	} else if (0 == strcmp("pragma", head->key)) {
		if (0 == strlen(head->val))
			return(0);
		head->type = HTTP_PRAGMA;
		if ( ! http_parse_pragma(head))
			return(0);
	} else if (0 == strcmp("content-type", head->key)) {
		if (0 == strlen(head->val))
			return(0);
		head->type = HTTP_CONTENT_TYPE;
		if ( ! http_parse_content_type(head))
			return(0);
	} else if (0 == strcmp("content-language", head->key)) {
		head->type = HTTP_CONTENT_LANG;
		head->data.clang = head->val;
	} else if (0 == strcmp("location", head->key)) {
		head->type = HTTP_LOCATION;
		if ( ! uri_parse(head->val, 0, &head->data.location))
			return(0);
		if ( ! uri_is_abs(&head->data.location))
			return(0);
	} else
		head->type = HTTP_UNKNOWN;

	return(1);
}

int
http_status(char *p, size_t sz, uint32_t *code)
{
	size_t		i, j;
	char		str[4];

	assert(NULL != code);
	assert(NULL != p);

	if (0 == sz)
		return(0);
	memset(str, 0, 4);

	for (i = 0; 32 != *p && i < sz; i++, p++)
		/* Skip to space. */ ;
	if (i++ == sz || 32 != *p++) {
		dddbg("%s: status-line bad syntax", __func__);
		return(0);
	}
	for (j = 0; j < 3 && i < sz; i++, j++)
		str[j] = *p++;
	if (i++ == sz || 32 != *p++) {
		dddbg("%s: status-line bad syntax", __func__);
		return(0);
	}

	*code = (uint32_t)atoi(str);
	return(1);
}

