/* $Id: resolv.c,v 1.10 2007-05-29 20:32:05 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>

#include <assert.h>
#include <err.h>
#include <getopt.h>
#include <inttypes.h>
#include <netdb.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <ares.h>

#include "nlpcrawl.h"


#define RESOLV_IP_WIDTH	(NLPCR_IP_WIDTH * sizeof(unsigned char))
#define RESOLV_TIMEO	(30)
#define RESOLV_MAXSIMUL	(64)
#define RESOLV_DNS_WAIT	(5)
#define RESOLV_CACHESZ	(4096)
#define RESOLV_QUEUESZ	(512)
#define RESOLV_DNS_AGE	(60 * 2)


struct dns_t {
	ares_channel	chan;
	size_t		pend;
	time_t		last;
};


struct ddns_t {
	struct dns_t	*dns;
	struct nlpcdbrp	*pair;
	void		*queue;
};


struct name_t {
	unsigned char	ip[NLPCR_IP_WIDTH];
	char		auth[NLPCR_URI_AUTHSZ];
	time_t		last;
	struct name_t	*prev;
	struct name_t	*next;
};


struct cache_t {
	void		*hash;
	struct name_t	*first;
	struct name_t	*last;
	size_t		max;
	size_t		size;
	size_t		age;
} cache;


static struct opt_t {
	void		*rqueue;
	void		*wqueue;
	size_t		pburst;
	size_t		pwait;
} opts;


static void
cache_free(void)
{
	struct name_t	*name, *p;

	name = cache.first;
	while (NULL != name) {
		p = name;
		name = name->next;
		free(p);
	}
	hash_free(&cache.hash);
	memset(&cache, 0, sizeof(struct cache_t));
}


static char *
auth_trim(const char in[NLPCR_URI_AUTHSZ], char out[NLPCR_URI_AUTHSZ])
{
	int		i;

	memset(out, 0, NLPCR_URI_AUTHSZ);
	for (i = 0; 0 != in[i] && ':' != in[i]; i++)
		out[i] = in[i];
	assert(i < NLPCR_URI_AUTHSZ);
	return(out);
}


static void
cache_rem(struct name_t *name)
{
	assert(NULL != name);
	assert(NULL != cache.first);
	assert(NULL != cache.last);

	if (NULL != name->next)
		name->next->prev = NULL == name->prev ?
			NULL : name->prev;
	if (NULL != name->prev)
		name->prev->next = NULL == name->next ?
			NULL : name->next;

	if (cache.first == name)
		cache.first = name->next;
	if (cache.last == name)
		cache.last = name->prev;
}


static int
cache_lookup(const struct nlpcdbrp *pair,
		unsigned char ip[NLPCR_IP_WIDTH])
{
	struct name_t	*name;
	char		buf[NLPCR_URI_AUTHSZ];
	time_t		cur;

	assert(NULL != pair);
	auth_trim(pair->record.uri.auth, buf);

	if ( ! hash_get(&cache.hash, buf, NLPCR_URI_AUTHSZ, 
				(void **)&name))
		return(0);

	memcpy(ip, name->ip, NLPCR_IP_WIDTH);
	time_s(cur);

	if (0 != cache.age && name->last + (time_t)cache.age < cur) {
		ddbg("%s: lookup expired: %s", __func__, buf);
		cache_rem(name);
		hash_del(&cache.hash, name->auth, NLPCR_URI_AUTHSZ);
		free(name);
		return(0);
	}

	time_s(name->last);

	assert(NULL != cache.first);
	assert(NULL != cache.last);

	cache_rem(name);
	if (NULL != cache.first)
		cache.first->prev = name;
	else
		cache.first = name;
	if (NULL == cache.last)
		cache.last = name;

	return(1);
}


static void
cache_add(const struct nlpcdbrp *pair)
{
	struct name_t	*name;
	char		buf[NLPCR_URI_AUTHSZ];

	assert(NULL != pair);
	auth_trim(pair->record.uri.auth, buf);

	if (hash_get(&cache.hash, buf, NLPCR_URI_AUTHSZ, NULL))
		return;

	++(cache.size);

	calloc_s(name, 1, sizeof(struct name_t));

	memcpy(name->ip, pair->record.ip, NLPCR_IP_WIDTH);
	memcpy(name->auth, buf, NLPCR_URI_AUTHSZ);
	time_s(name->last);

	if (NULL != cache.first) {
		assert(NULL != cache.last);
		name->next = cache.first;
		cache.first->prev = name;
		cache.first = name;
	} else {
		assert(NULL == cache.last);
		cache.first = cache.last = name;
	}

	hash_put(&cache.hash, buf, NLPCR_URI_AUTHSZ, name);

	if (0 != cache.max && cache.size > cache.max) {
		ddbg("%s: cache exceeded: %zd",
				__func__, cache.max);
		name = cache.last;
		if (NULL != name->prev)
			name->prev->next = NULL;
		cache.last = name->prev;
		hash_del(&cache.hash, name->auth, NLPCR_URI_AUTHSZ);
		free(name);
	}
}


void
dns_cb(void *arg, int status, struct hostent *hosts)
{
	struct ddns_t	*data;
	char		tmp[NLPCR_IP_STRING],
			buf[NLPCR_URI_FMTSZ];
	unsigned char	*addr;
	uint64_t	flags;
	struct nlpcdbrp	*pair;
	struct dns_t	*dns;

	assert(NULL != arg);
	data = (struct ddns_t *)arg;
	assert(NULL != data->dns);
	dns = data->dns;
	assert(NULL != data->pair);
	pair = data->pair;
	
	flags = SQUEUE_WAIT | SQUEUE_NOTIFY;
	--(dns->pend);
	time_s(dns->last);

	if (ARES_SUCCESS != status) {
		pair->record.state = NLPCDBR_UNREACH;
		pair->record.code = NLPCDBR_RESOLV;

		ddbg("%s: fail: lookup failed: %s, %s",
			__func__,  ares_strerror(status),
			uri_topstring_r(&pair->record.uri, buf));
		squeue_add(opts.wqueue, pair, flags);
		data->dns = NULL;
		data->pair = NULL;
		q_add(&data->queue, data);
		return;
	} 

	addr = (unsigned char *)hosts->h_addr_list[0];
	memcpy(pair->record.ip, addr, RESOLV_IP_WIDTH);
	pair->record.state = NLPCDBR_READY;
	pair->record.code = NLPCDBR_OK;

	dbg("lookup: %s, %s", 
		ip_tostring_r(pair->record.ip, tmp), 
		uri_topstring_r(&pair->record.uri, buf));

	squeue_add(opts.wqueue, pair, flags);
	cache_add(pair);
	data->dns = NULL;
	data->pair = NULL;
	q_add(&data->queue, data);
}


static void
dns_init(struct dns_t *p)
{
	int		aflags, rc;
	struct ares_options	aopt;

	assert(NULL != p);

	aopt.flags = ARES_FLAG_NOSEARCH;
	aopt.timeout = RESOLV_TIMEO;
	aflags = ARES_OPT_FLAGS | ARES_OPT_TIMEOUT;

	rc = ares_init_options(&p->chan, &aopt, aflags);
	if (ARES_SUCCESS != rc)
		diex("%s: fail: ares_init_options: %s", __func__,
				ares_strerror(rc));
	p->pend = 0;
	p->last = 0;
}


static int
dns_assign(struct dns_t *p, void *dataq, struct nlpcdbrp *pair)
{
	char		buf[NLPCR_URI_AUTHSZ];
	struct ddns_t	*data;

	assert(NULL != p);
	assert(NULL != pair);
	assert(NULL != dataq);

	if (p->pend >= opts.pburst)
		return(0);

	auth_trim(pair->record.uri.auth, buf);

	++(p->pend);

	data = (struct ddns_t *)q_rem(&dataq);
	assert(NULL != data);

	data->dns = p;
	data->pair = pair;
	ares_gethostbyname(p->chan, buf, AF_INET, dns_cb, data);

	ddbg("%s: scheduling: [%s]", __func__, buf);

	return(1);
}


static struct nlpcdbrp *
dns_next(uint64_t fl)
{
	char		buf[NLPCR_URI_FMTSZ],
			bufip[NLPCR_IP_STRING];
	unsigned char	ip[NLPCR_IP_WIDTH];
	struct nlpcdbrp	*p;

	for (;;) {
		p = (struct nlpcdbrp *)squeue_rem(opts.rqueue, fl);
		if (NULL == p)
			break;

		assert(0 != p->record.uri.auth[0]);
		if ( ! cache_lookup(p, ip))
			break;
		memcpy(p->record.ip, ip, NLPCR_IP_WIDTH);
		p->record.code = NLPCDBR_OK;
		p->record.state = NLPCDBR_READY;
		dbg("cache-hit: %s, %s", 
				ip_tostring_r(p->record.ip, bufip),
				uri_topstring_r(&p->record.uri, buf));
		squeue_add(opts.wqueue, p, 
				SQUEUE_NOTIFY | SQUEUE_WAIT);
	}
	return(p);
}


void *
thread_exec(void *arg)
{
	struct dns_t	dns;
	struct ddns_t	*ddns;
	void		*dq;
	struct nlpcdbrp	*pair;
	uint64_t	wflg, nwflg;
	int		fdn;
	size_t		i;
	time_t		cur;
	struct timeval	tv;
	fd_set		fdsr, fdsw;

	assert(NULL == arg);

	wflg = SQUEUE_NOTIFY | SQUEUE_WAIT;
	nwflg = SQUEUE_NOTIFY;

	calloc_s(ddns, opts.pburst, sizeof(struct ddns_t));
	dns_init(&dns);
	q_new(&dq);

	for (i = 0; i < opts.pburst; i++) {
		ddns[i].queue = dq;
		q_add(&dq, &ddns[i]);
	}

	pthread_cleanup_push((void(*)(void*))q_free, (void *)&dq);
	pthread_cleanup_push((void(*)(void*))ares_destroy, 
			(void *)dns.chan);
	pthread_cleanup_push(free, (void *)ddns);

	for (;;) {
		FD_ZERO(&fdsr);
		FD_ZERO(&fdsw);

		pthread_testcancel();

		if (0 != (fdn = ares_fds(dns.chan, &fdsr, &fdsw))) {
			memset(&tv, 0, sizeof(struct timeval));
			ares_timeout(dns.chan, NULL, &tv);

			if (-1 == select(fdn, &fdsr, &fdsw, NULL, &tv))
				die("%s: fail: select", __func__);

			ares_process(dns.chan, &fdsr, &fdsw);
		}

		time_s(cur);
		if (dns.last + (time_t)opts.pwait > cur) {
			if (0 != dns.pend)
				continue;
			sleep((dns.last + opts.pwait) - cur);
		}

		pair = dns_next(0 == dns.pend ? wflg : nwflg);
		if (NULL == pair)
			continue;
		while (dns_assign(&dns, dq, pair))
			if (NULL == (pair = dns_next(nwflg)))
				break;
	}

	/* NOTREACHED */
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);

	return(NULL);
}


void *
thread_write(void *arg)
{
	struct nlpcdbrp	*p;
	struct uri	*addr;
	char		buf[NLPCR_URI_FMTSZ], reason[32];
	size_t		u, r;
	uint64_t	wflags;

	assert(NULL == arg);

	wflags = SQUEUE_WAIT | SQUEUE_NOTIFY;
	u = r = 0;

	for (;;) {
		p = squeue_rem(opts.wqueue, SQUEUE_NOTIFY);
		if (NULL == p) {
			if (u > 0) 
				signal_next(NLPCDBR_UNREACH);
			if (r > 0)
				signal_next(NLPCDBR_READY);
			u = r = 0;
			p = squeue_rem(opts.wqueue, wflags);
			assert(NULL != p);
		}

		addr = &p->record.uri;
		uri_topstring_r(addr, buf);
		nlpcdbr_code2str_r(reason, p->record.code);

		if ( ! db_put(p))
			diex("%s: fail: db_put: %s", __func__, buf);

		switch (p->record.state) {
		case (NLPCDBR_UNREACH):
			u++;
			dbg("accept: unreach (%s): %s", reason, buf);
			break;
		case (NLPCDBR_READY):
			r++;
			dbg("accept: ready: %s", buf);
			break;
		default:
			abort();
		}
		free(p);
	}

	/* NOTREACHED */
	return(NULL);
}


static int
scan(void)
{
	struct nlpcdbrp	*p;
	struct db_query	query;
	void		*cookie;
	int		flags, rc;
	static uint64_t	last;

	memset(&query, 0, sizeof(struct db_query));

	query.field = DB_FIELD_STATE;
	query.state = NLPCDBR_NEW | NLPCDBR_DIRTY;
	query.skip = last;

	db_window_fill(&query);
	db_query_begin(&cookie);

	for (rc = 1, p = NULL, flags = 0; ; ) {
		if (SIGNAL_EXIT & (flags |= signal_test())) 
			break;
		calloc_s(p, 1, sizeof(struct nlpcdbrp));
		if (1 != (rc = db_query_next(&cookie, &query, p)))
			break;
		if ( ! squeue_add(opts.rqueue, p, SQUEUE_NOTIFY))
			break;
		p = NULL;
	}

	db_query_close(&cookie);
	last = (0 == rc ? 0 : query.pos);
	if (NULL != p)
		free(p);

	return(flags);
}


int
main(int argc, char *argv[])
{
	int 		c, t;
	char		*dir;

	extern char	*__progname;
	extern char 	*optarg;
	extern int 	optind;
	pthread_t	execer, writer;

	static struct option longopts[] = {
		{ "fetch-burst",	required_argument, NULL, 1 },
		{ "fetch-wait",		required_argument, NULL, 2 },
		{ "cache-size",		required_argument, NULL, 3 },
		{ "cache-maxage",	required_argument, NULL, 4 },
		{ NULL, 0, NULL, 0 }
	};

	dir = NLPC_DEFAULT_DIR;

	memset(&opts, 0, sizeof(struct opt_t));
	opts.pburst = RESOLV_MAXSIMUL;
	opts.pwait = RESOLV_DNS_WAIT;

	memset(&cache, 0, sizeof(struct cache_t));
	cache.max = RESOLV_CACHESZ;
	cache.age = RESOLV_DNS_AGE;
	hash_new(&cache.hash);

	while (-1 != (c = getopt_long(argc, argv, "hvdp:",
					longopts, NULL)))
		switch (c) {
		case (1):
			if ( ! parse_int(optarg, &t, INT_MAX, 1))
				diex("invalid value: --fetch-burst");
			opts.pburst = (size_t)t;
			break;
		case (2):
			if ( ! parse_int(optarg, &t, INT_MAX, 1))
				diex("invalid value: --fetch-wait");
			opts.pwait = (size_t)t;
			break;
		case (3):
			if ( ! parse_int(optarg, &t, INT_MAX, 0))
				diex("invalid value: --cache-size");
			cache.max = (size_t)t;
			break;
		case (4):
			if ( ! parse_int(optarg, &t, INT_MAX, 0))
				diex("invalid value: --cache-maxage");
			cache.age = (size_t)t;
			break;
		case ('d'):
			dbg_lvl++;
			break;
		case ('p'):
			dir = optarg;
			break;
		case ('v'):
			fprintf(stdout, "%s version %s\n", 
					__progname, VERSION);
			return(EXIT_SUCCESS);
		case ('h'):
			/*FALLTHROUGH*/
		default:
			fprintf(stderr, "usage: %s [-dhv] [-p path] "
					"[LONG_OPTS]\n", __progname);
			return(EXIT_FAILURE);
		}
	argc -= optind;
	argv += optind;

	opts.wqueue = squeue_new(RESOLV_QUEUESZ);
	opts.rqueue = squeue_new(RESOLV_QUEUESZ);

	signal_init(dir, NLPCDBR_NEW | NLPCDBR_DIRTY);
	db_init(dir, 0);

	if (0 != pthread_create(&execer, NULL, thread_exec, NULL))
		die("%s: fail: pthread_create", __func__);
	if (0 != pthread_create(&writer, NULL, thread_write, NULL))
		die("%s: fail: pthread_create", __func__);

	for (;;) {
		if (SIGNAL_EXIT & scan())
			break;
		if (SIGNAL_EXIT & signal_wait(0))
			break;
	}

	if (0 != pthread_cancel(execer))
		die("%s: fail: pthread_cancel", __func__);
	if (0 != pthread_cancel(writer))
		die("%s: fail: pthread_cancel", __func__);
	if (0 != pthread_join(execer, NULL))
		die("%s: fail: pthread_join", __func__);
	if (0 != pthread_join(writer, NULL))
		die("%s: fail: pthread_join", __func__);

	db_teardown();
	signal_teardown();

	cache_free();

	squeue_free(opts.wqueue);
	squeue_free(opts.rqueue);

	return(EXIT_SUCCESS);
}

