/* $Id: signal.c,v 1.29 2007-05-27 20:28:20 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>
#ifdef __linux__
# include <sys/file.h>
#endif

#include <assert.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "nlpcrawl.h"


/*
 * Maximum length of contents in a pidfile.
 */
#define PIDFILE_SIZE	(64)


/*
 * The on-disk pidfile (only valid if running in daemon mode).
 */
static char		pidfile[MAXPATHLEN] = {0};

/*
 * The prefix for all pidfiles.
 */
static char		piddir[MAXPATHLEN] = {0};


int
signal_test(void)
{
	sigset_t	set;
	int		flags, p, sig;

	flags = 0;
	if (-1 == sigemptyset(&set))
		die("fail: sigpending");
	if (-1 == sigpending(&set))
		die("fail: sigpending");
	for (;;) {
		p = 0;
		if (sigismember(&set, NLPC_SIGNAL_EUP)) {
			p++;
			dddbg("%s: eup request", __func__);
			db_window_grow();
		}
		if (sigismember(&set, NLPC_SIGNAL_EDOWN)) {
			p++;
			dddbg("%s: edown request", __func__);
			db_window_shrink();
		}
		if (sigismember(&set, NLPC_SIGNAL_TERM)) {
			p++;
			dddbg("%s: term request", __func__);
			flags |= SIGNAL_EXIT;
		}
		if (sigismember(&set, NLPC_SIGNAL_SCAN)) {
			p++;
			dddbg("%s: scan request", __func__);
			flags |= SIGNAL_SCAN;
		}
		if (0 == p)
			break;
		if (-1 == sigwait(&set, &sig))
			die("fail: sigwait");
		if (-1 == sigemptyset(&set))
			die("fail: sigpending");
		if (-1 == sigpending(&set))
			die("fail: sigpending");
	}
	return(flags);
}


int
signal_wait(unsigned int sec)
{
	int		flags;

	dddbg("%s: %u", __func__, sec);
	if (0 == sec) {
		dddbg("%s: (blocking)", __func__);
		sec = UINT_MAX;
	}

	for (;;) {
		sleep(1);
		if (--sec == 0)
			return(SIGNAL_TIMEOUT);
		else if (0 != (flags = signal_test()))
			break;
	}
	return(flags);
}


static char *
signal_file(char *path, uint64_t state)
{
	char		*sstate;

	assert(NULL != path);
	memcpy(path, piddir, MAXPATHLEN);
	assert(0 == path[MAXPATHLEN - 2]);

	if ('/' != path[strlen(path) - 1]) {
		strncat(path, "/", MAXPATHLEN - strlen(path));
		assert (0 == path[MAXPATHLEN - 1]);
	}

	if (state & NLPCDBR_NEW) 
		sstate = "nlpc-resolv";
	else if (state & NLPCDBR_DIRTY)
		sstate = "nlpc-resolv";
	else if (state & NLPCDBR_FRESH)
		sstate = "nlpc-vrfy";
	else if (state & NLPCDBR_READY)
		sstate = "nlpc-fetch";
	else if (state & NLPCDBR_UNREACH)
		sstate = "nlpc-up";
	else if (state & NLPCDBR_DEAD)
		sstate = "nlpc-up";
	else if (state & NLPCDBR_SCAN)
		sstate = "nlpc-up";
	else if (state & NLPCDBR_VRFY)
		sstate = "nlpc-scan";
	else
		abort();

	strncat(path, sstate, MAXPATHLEN - strlen(path));
	assert(0 == path[MAXPATHLEN - 1]);

	return(path);
}


void
signal_init(const char *prefix, uint64_t state)
{
	int		fd, rc;
	char		tmp[MAXPATHLEN], contents[PIDFILE_SIZE],
			path[MAXPATHLEN];
	sigset_t	mask;
	size_t		sz;

	assert(0 == pidfile[0]);
	assert(0 == piddir[0]);
	assert(NULL != prefix);

	/*
	 * Carefully make sure that we don't overrun any particular part
	 * of the buffer in this sequence.
	 */

	if (0 == prefix[0])
		diex("fail: prefix is zero-length");

	strncat(piddir, prefix, MAXPATHLEN);
	if (0 != piddir[MAXPATHLEN - 2])
		diex("fail: prefix too long: %s", prefix);

	if ('/' != piddir[strlen(piddir) - 1]) {
		strncat(piddir, "/", MAXPATHLEN - strlen(piddir));
		assert(0 == piddir[MAXPATHLEN - 1]);
	}

	strncat(piddir, "run", MAXPATHLEN - strlen(piddir));
	if (0 != piddir[MAXPATHLEN - 1])
		diex("fail: prefix too long: %s", prefix);
	
	/* Note that mkdir_p modifies the underlying buffer, so use a
	 * copy in order to construct the hierarchy. */

	memcpy(tmp, piddir, MAXPATHLEN);
	if (mkdir_p(tmp, 0755))
		die("fail: mkdir_p: %s", piddir);

	dddbg("%s: initialised: %s", __func__, piddir);

	/*
	 * Now block the non-fatal signals.  They'll be atomically
	 * unblocked during signal_wait (we only care to receive them
	 * during the wait/check period).  Note that if we've no
	 * classification, simply ignore signals and exit.
	 */

	if (0 == state) {
		if (SIG_ERR == signal(NLPC_SIGNAL_SCAN, SIG_IGN))
			die("%s: fail: signal", __func__);
		if (SIG_ERR == signal(NLPC_SIGNAL_EUP, SIG_IGN))
			die("%s: fail: signal", __func__);
		if (SIG_ERR == signal(NLPC_SIGNAL_EDOWN, SIG_IGN))
			die("%s: fail: signal", __func__);
		if (SIG_ERR == signal(NLPC_SIGNAL_TERM, SIG_IGN))
			die("%s: fail: signal", __func__);
		return;
	}

	if (-1 == sigemptyset(&mask))
		die("fail: sigemptyset");
	if (-1 == sigaddset(&mask, NLPC_SIGNAL_SCAN))
		die("fail: sigaddset");
	if (-1 == sigaddset(&mask, NLPC_SIGNAL_EUP))
		die("fail: sigaddset");
	if (-1 == sigaddset(&mask, NLPC_SIGNAL_EDOWN))
		die("fail: sigaddset");
	if (-1 == sigaddset(&mask, NLPC_SIGNAL_TERM))
		die("fail: sigaddset");
	if (-1 == sigprocmask(SIG_BLOCK, &mask, NULL))
		die("fail: sigprocmask");

	dddbg("%s: signals blocked: %d, %d, %d, %d", __func__,
			NLPC_SIGNAL_SCAN, NLPC_SIGNAL_TERM,
			NLPC_SIGNAL_EUP, NLPC_SIGNAL_EDOWN);

	/*
	 * Construct our pidfile reference, then open and lock it, and
	 * write our magic to it.
	 */

	strncpy(pidfile, signal_file(path, state), MAXPATHLEN);

	memset(contents, 0, PIDFILE_SIZE);
	snprintf(contents, PIDFILE_SIZE, "%d", getpid());
	assert(0 == contents[PIDFILE_SIZE - 1]);

	sz = strlen(contents);
	rc = EXIT_SUCCESS;

	if (-1 == (fd = open(pidfile, O_WRONLY | O_CREAT |
					O_EXCL, 0644)))
		die("fail: open: %s", pidfile);

	if (-1 == flock(fd, LOCK_EX)) {
		warn("fail: lock: %s", pidfile);
		rc = EXIT_FAILURE;
	} else if ((ssize_t)sz != write(fd, contents, sz)) {
		warn("fail: write: %s", pidfile);
		rc = EXIT_FAILURE;
	}

	if (-1 == flock(fd, LOCK_UN)) {
		warn("fail: funlock: %s", pidfile);
		rc = EXIT_FAILURE;
	}

	if (EXIT_SUCCESS == rc) {
		dddbg("%s: initialised: %s", __func__, pidfile);
		return;
	}

	if (-1 == close(fd))
		warn("fail: close: %s", pidfile);
	if (-1 == remove(pidfile))
		warn("fail: remove: %s", pidfile);

	exit(EXIT_FAILURE);
	/* NOTREACHED */
}


void
signal_teardown(void)
{
	if (0 == pidfile[0]) {
		dddbg("%s: torn down", __func__);
		return;
	}
	if (-1 == remove(pidfile))
		ddbg("warn: remove: %s", pidfile);
	dddbg("%s: torn down: %s", __func__, pidfile);
}


int
signal_next(uint64_t state)
{
	int		fd, errn, rc;
	char		contents[PIDFILE_SIZE], *pf,
			path[MAXPATHLEN], errs[255];
	pid_t		pid;

	memset(contents, 0, PIDFILE_SIZE);

	pf = signal_file(path, state);
	if (-1 == (fd = open(pf, O_RDONLY))) {
		if (ENOENT != (errn = errno))
			ddbg("warn: open: %s, %s", pf, strerror(errn));
		return(0);
	}

	rc = 1;

	/*
	 * Establish a shared (non-exclusive) lock on the target
	 * pidfile, read its contents, then try to kill it.
	 */

	if (-1 == flock(fd, LOCK_SH)) {
		errn = errno;
		strerror_r(errn, errs, 255);
		ddbg("%s: warn: flock: %s, %s", __func__, pf, errs);
		rc = 0;
	} else if (-1 == read(fd, contents, PIDFILE_SIZE - 1)) {
		errn = errno;
		strerror_r(errn, errs, 255);
		ddbg("%s: warn: read: %s, %s", __func__, pf, errs);
		rc = 0;
	}

	if (-1 == flock(fd, LOCK_UN)) {
		errn = errno;
		strerror_r(errn, errs, 255);
		ddbg("%s: warn: funlock: %s, %s", __func__, pf, errs);
		rc = 0;
	}
	if (-1 == close(fd)) {
		errn = errno;
		strerror_r(errn, errs, 255);
		ddbg("%s: warn: close: %s, %s", __func__, pf, errs);
		rc = 0;
	}

	if (0 == rc)
		return(0);

	if ( ! parse_int(contents, &pid, INT_MAX, 1)) {
		ddbg("%s: warn: parse integer: %s", __func__, pf);
		return(0);
	}

	if (0 == kill(pid, NLPC_SIGNAL_SCAN)) {
		ddbg("%s: signal-scan: %d, %s", __func__, pid, pf);
		return(1);
	}

	dddbg("%s: notified: %s", __func__, pf);
	return(0);
}

