/* $Id: test.c,v 1.8 2007-04-23 01:41:46 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>

#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <check.h>

#include "nlpcrawl.h"

#define HTTP_PARSE_MAXLINE (4096)

static char *
uri2str(const char *prefix, const char p[NLPCR_URI_MAXSZ])
{
	static char	msg[NLPCR_URI_MAXSZ * 2];
	size_t		ssz;
	
	ssz = (NLPCR_URI_MAXSZ * 2 - 1);

	memset(msg, 0, sizeof(msg));
	strncat(msg, prefix, ssz);
	strncat(msg, ": [", ssz);
	strncat(msg, p, ssz);
	strncat(msg, "]", ssz);

	return(msg);
}

static char *
query2str(const char *prefix, const char p[HTTP_PARSE_MAXLINE])
{
	static char 	msg[HTTP_PARSE_MAXLINE * 2];
	size_t		i, sz, ssz;

	sz = strlen(p);
	ssz = (HTTP_PARSE_MAXLINE * 2 - 1);

	memset(msg, 0, sizeof(msg));

	strncat(msg, prefix, ssz);
	strncat(msg, ": [", ssz);
	for (i = 0; i < sz; i++)
		switch(p[i]) {
		case ('\r'):
			strncat(msg, "\\r", ssz);
			break;
		case ('\n'):
			strncat(msg, "\\n", ssz);
			break;
		case ('\t'):
			strncat(msg, "\\t", ssz);
			break;
		default:
			strncat(msg, &p[i], 1);
			break;
		}
	strncat(msg, "]", ssz);
	return(msg);
}

START_TEST(uri_abs)
{
#define URI_PASS_ABS	(11)
#define URI_FAIL_ABS	(2)
	int 		i;
	struct uri 	p;

	struct {
		char line[NLPCR_URI_MAXSZ];
		char expect[NLPCR_URI_MAXSZ];
	} pass[URI_PASS_ABS] = {
		{ "g:", 
		  "g:" },
		{ "G:", 
		  "g:" },
		{ "g:h", 
		  "g:h" },
		{ "G:H", 
		  "g:H" },
		{ "g:a/b/c", 
		  "g:a/b/c" },
		{ "g:A/B/C", 
		  "g:A/B/C" },
		{ "g://a/b/c", 
		  "g://a/b/c" },
		{ "g://A/b/c", 
		  "g://a/b/c" },
		{ "g://a:80/b/c", 
		  "g://a:80/b/c" },
		{ "g://a:80/b/c   ", 
		  "g://a:80/b/c" },
		{ "g://a:80/b/c       ", 
		  "g://a:80/b/c" },
	};

	char fail[URI_FAIL_ABS][NLPCR_URI_MAXSZ] = {
		{ "" },
		{ "g" },
	};

	for (i = 0; i < URI_PASS_ABS; i++) {
		fail_unless(uri_parse_abs(pass[i].line, 0, &p),
			uri2str("fail: uri", pass[i].line));
		fail_unless(0 == strcmp(uri_tostring(&p), 
					pass[i].expect),
			uri2str("fail: uri", pass[i].line));
	}

	for (i = 0; i < URI_FAIL_ABS; i++) 
		fail_unless( ! uri_parse_abs(fail[i], 0, &p),
			uri2str("fail-ok: uri", fail[i]));
} END_TEST

START_TEST(http_status_general)
{
#define HTTP_PASS_STATUS (3)
#define HTTP_FAIL_STATUS (2)
	int 		i;
	uint32_t 	code;

	struct {
		char line[HTTP_PARSE_MAXLINE];
		char save[HTTP_PARSE_MAXLINE];
		uint32_t expect;
	} pass[HTTP_PASS_STATUS] = {
		{ "HTTP/1.1 200 OK", "HTTP/1.1 200 OK", 200 },
		{ "HTTP/1.1 200 OK\r\n", "HTTP/1.1 200 OK\r\n", 200 },
		{ "HTTP/1.0 200 SOMETHING HERE",
		  "HTTP/1.0 200 SOMETHING HERE", 200 }
	};

	struct {
		char line[HTTP_PARSE_MAXLINE];
		char save[HTTP_PARSE_MAXLINE];
	} fail[HTTP_FAIL_STATUS] = {
		{ "HTTP/1.1\r\n200 OK", "HTTP/1.1\r\n200 OK" },
		{ "HTTP/1.1 200\r\nOK", "HTTP/1.1 200\r\nOK" }
	};

	for (i = 0; i < HTTP_PASS_STATUS; i++) {
		fail_unless(http_status(pass[i].line, 
				strlen(pass[i].line), &code), 
			query2str("fail: status-line", pass[i].save));
		fail_unless(code == pass[i].expect,
			query2str("fail: status-line", pass[i].save));
	}
	for (i = 0; i < HTTP_FAIL_STATUS; i++)
		fail_unless( ! http_status(fail[i].line, 
					strlen(fail[i].line), &code), 
			query2str("fail-ok: status-line", 
				fail[i].save));
} END_TEST

START_TEST(http_parse_location)
{
#define HTTP_PASS_LOCATION (3)
#define HTTP_FAIL_LOCATION (3)
	struct http_header h;
	int 		i;

	struct {
		char line[HTTP_PARSE_MAXLINE];
		char save[HTTP_PARSE_MAXLINE];
		struct uri expect;
	} pass[HTTP_PASS_LOCATION] = {
		{ "Location: http://a/b/c/d\r\n",
		  "Location: http://a/b/c/d\r\n",
		  { "http", "a", "/b/c/d", "", "" }},
		{ "Location:http://a/b/c/d\r\n",
		  "Location:http://a/b/c/d\r\n",
		  { "http", "a", "/b/c/d", "", "" }},
		{ "Location:\r\n  http://a/b/c/d?e\r\n",
		  "Location:\r\n  http://a/b/c/d?e\r\n",
		  { "http", "a", "/b/c/d", "e", "" }}
	};

	struct {
		char line[HTTP_PARSE_MAXLINE];
		char save[HTTP_PARSE_MAXLINE];
	} fail[HTTP_FAIL_LOCATION] = {
		{ "Location: a/b/c/d\r\n", "Location: a/b/c/d\r\n" },
		{ "Location: \r\n", "Location: \r\n" },
		{ "Location:", "Location:" }
	};

	for (i = 0; i < HTTP_PASS_LOCATION; i++) {
		fail_unless(http_parse(pass[i].line, 
				strlen(pass[i].line), &h), 
			query2str("fail: location", pass[i].save));
		fail_unless(HTTP_LOCATION == h.type,
			query2str("fail: location", pass[i].save));
	}
	for (i = 0; i < HTTP_FAIL_LOCATION; i++)
		fail_unless( ! http_parse(fail[i].line, 
					strlen(fail[i].line), &h), 
			query2str("fail-ok: location", fail[i].save));
} END_TEST

START_TEST(http_parse_content_language)
{
#define HTTP_PASS_CLANG (3)
#define HTTP_FAIL_CLANG (3)
	struct http_header h;
	int 		i;

	struct {
		char line[HTTP_PARSE_MAXLINE];
		char save[HTTP_PARSE_MAXLINE];
		char expect[HTTP_PARSE_MAXLINE];
	} pass[HTTP_PASS_CLANG] = {
		{ "Content-Language: lv\r\n",
		  "Content-Language: lv\r\n", "lv" },
		{ "Content-Language:lv\r\n",
		  "Content-Language:lv\r\n", "lv" },
		{ "Content-Language:\r\nen_GB\r\n",
		  "Content-Language:\r\nen_GB", "en_GB" }
	};

	struct {
		char line[HTTP_PARSE_MAXLINE];
		char save[HTTP_PARSE_MAXLINE];
	} fail[HTTP_FAIL_CLANG] = {
		{ "Content-Language: ", "Content-Language: " },
		{ "Content-Language:", "Content-Language:" },
		{ "Content-Language:\r\n", "Content-Language:\r\n" }
	};

	for (i = 0; i < HTTP_PASS_CLANG; i++) {
		fail_unless(http_parse(pass[i].line, 
				strlen(pass[i].line), &h), 
			query2str("fail: language", pass[i].save));
		fail_unless(HTTP_CONTENT_LANG == h.type,
			query2str("fail: language", pass[i].save));
		fail_unless(0 == strcmp(h.data.clang, 
					pass[i].expect),
			query2str("fail: language", pass[i].save));
	}
	for (i = 0; i < HTTP_FAIL_CLANG; i++)
		fail_unless( ! http_parse(fail[i].line, 
					strlen(fail[i].line), &h), 
			query2str("fail-ok: language", fail[i].save));
} END_TEST

START_TEST(http_parse_content_type)
{
#define HTTP_PASS_CTYPE (7)
#define HTTP_FAIL_CTYPE (5)
	int 		i;
	struct http_header h;

	struct {
		char line[HTTP_PARSE_MAXLINE];
		char save[HTTP_PARSE_MAXLINE];
		struct http_header_content_type expect;
	} pass[HTTP_PASS_CTYPE] = {
		{ "Content-Type: text/xml;charset=utf-8\r\n",
		  "Content-Type: text/xml;charset=utf-8\r\n",
			{ NLPCDBR_TYPE_TXT_XML, "utf-8" }},
		{ "Content-Type: text/xml;   charset=utf-8\r\n",
		  "Content-Type: text/xml;   charset=utf-8\r\n",
			{ NLPCDBR_TYPE_TXT_XML, "utf-8" }},
		{ "Content-Type:text/html;\r\ncharset=utf-8\r\n",
		  "Content-Type:text/html;\r\ncharset=utf-8\r\n",
			{ NLPCDBR_TYPE_TXT_HTML, "utf-8" }},
		{ "Content-Type:text/html\r\n\r\n",
		  "Content-Type:text/html\r\n\r\n",
			{ NLPCDBR_TYPE_TXT_HTML, "" }},
		{ "Content-Type:     application/xml;charset=utf-8",
		  "Content-Type:     application/xml;charset=utf-8",
			{ NLPCDBR_TYPE_APP_XML, "utf-8" }},
		{ "Content-Type:     application/xhtml+html",
		  "Content-Type:     application/xhtml+html",
			{ NLPCDBR_TYPE_APP_XHTML, "" }},
		{ "Content-Type:\r\ntext/plain;\r\ncharset=utf-8",
		  "Content-Type:\r\ntext/plain;\r\ncharset=utf-8",
			{ NLPCDBR_TYPE_TXT_PLAIN, "utf-8" }}
	};

	struct {
		char line[HTTP_PARSE_MAXLINE];
		char save[HTTP_PARSE_MAXLINE];
	} fail[HTTP_FAIL_CTYPE] = {
		{ "Content-Type:\r\n   ", "Content-Type:\r\n   " },
		{ "Content-Type:  \r\n   ", "Content-Type:  \r\n   " },
		{ "Content-Type:", "Content-Type:" },
		{ "Content-Type: ;charset=utf-8", 
		  "Content-Type: ;charset=utf-8"},
		{ "Content-Type: text/html;charset=", 
		  "Content-Type: text/html;charset="},
	};

	for (i = 0; i < HTTP_PASS_CTYPE; i++) {
		fail_unless(http_parse(pass[i].line, 
				strlen(pass[i].line), &h), 
			query2str("fail: ctype", pass[i].save));
		fail_unless(HTTP_CONTENT_TYPE == h.type,
			query2str("fail: ctype", pass[i].save));
		fail_unless(h.data.ctype.media ==
				pass[i].expect.media,
			query2str("fail: ctype", pass[i].save));
		if (0 == strlen(pass[i].expect.charset))
			continue;
		fail_unless(0 == strcmp(h.data.ctype.charset,
					pass[i].expect.charset),
			query2str("fail: ctype", pass[i].save));

	}
	for (i = 0; i < HTTP_FAIL_CTYPE; i++)
		fail_unless( ! http_parse(fail[i].line, 
					strlen(fail[i].line), &h),
			query2str("fail-ok: ctype", fail[i].save));
} END_TEST

START_TEST(http_parse_pragma)
{
#define HTTP_PASS_PRAGMA (6)
#define HTTP_FAIL_PRAGMA (3)
	int 		i;
	struct http_header h;

	struct {
		char line[HTTP_PARSE_MAXLINE];
		char save[HTTP_PARSE_MAXLINE];
		struct http_header_pragma expect;
	} pass[HTTP_PASS_PRAGMA] = {
		{ "Pragma: No-Cache\r\n", "Pragma: No-Cache\r\n", 
			{ HTTP_PRAGMA_NOCACHE, {1}}},
		{ "Pragma: Cache\r\n", "Pragma: Cache\r\n", 
			{ HTTP_PRAGMA_UNKNOWN, {0}}},
		{ "Pragma: No-Cache", "Pragma: No-Cache", 
			{ HTTP_PRAGMA_NOCACHE, {1}}},
		{ "Pragma: Cache", "Pragma: Cache", 
			{ HTTP_PRAGMA_UNKNOWN, {0}}},
		{ "Pragma:     No-Cache ", "Pragma:     No-Cache ", 
			{ HTTP_PRAGMA_NOCACHE, {1}}},
		{ "Pragma:     Cache ", "Pragma:     Cache ", 
			{ HTTP_PRAGMA_UNKNOWN, {0}}},
	};

	struct {
		char line[HTTP_PARSE_MAXLINE];
		char save[HTTP_PARSE_MAXLINE];
	} fail[HTTP_FAIL_PRAGMA] = {
		{ "Pragma:\r\n   ", "Pragma:\r\n   " },
		{ "Pragma:  \r\n   ", "Pragma:  \r\n   " },
		{ "Pragma:", "Pragma:" }
	};

	for (i = 0; i < HTTP_PASS_PRAGMA; i++) {
		fail_unless(http_parse(pass[i].line, 
				strlen(pass[i].line), &h), 
			query2str("fail: pragma", pass[i].save));
		fail_unless(HTTP_PRAGMA == h.type,
			query2str("fail: pragma", pass[i].save));
		fail_unless(h.data.pragma.pragma_type ==
				pass[i].expect.pragma_type,
			query2str("fail: pragma", pass[i].save));
		fail_unless(h.data.pragma.data.nocache ==
				pass[i].expect.data.nocache,
			query2str("fail: pragma", pass[i].save));
	}
	for (i = 0; i < HTTP_FAIL_PRAGMA; i++)
		fail_unless( ! http_parse(fail[i].line, 
					strlen(fail[i].line), &h),
			query2str("fail-ok: pragma", fail[i].save));
} END_TEST

START_TEST(http_parse_expires)
{
#define HTTP_PASS_EXPIRES (3)
#define HTTP_FAIL_EXPIRES (2)
	struct http_header h;
	int 		i;

	struct {
		char line[HTTP_PARSE_MAXLINE];
		char save[HTTP_PARSE_MAXLINE];
		char expect[HTTP_PARSE_MAXLINE];
	} pass[HTTP_PASS_EXPIRES] = {
		{ "Expires: Mon, 09 Apr 2007 20:39:36 GMT\r\n",
		  "Expires: Mon, 09 Apr 2007 20:39:36 GMT\r\n",
		  "Mon, 09 Apr 2007 20:39:36 GMT" },
		{ "Expires:      Mon, 09 Apr 2007 20:39:36 GMT  \r\n",
		  "Expires:      Mon, 09 Apr 2007 20:39:36 GMT  \r\n",
		  "Mon, 09 Apr 2007 20:39:36 GMT" },
		{ "Expires: Mon, 09 Apr 2007 20:39:36 GMT",
		  "Expires: Mon, 09 Apr 2007 20:39:36 GMT",
		  "Mon, 09 Apr 2007 20:39:36 GMT" },
	};

	struct {
		char line[HTTP_PARSE_MAXLINE];
		char save[HTTP_PARSE_MAXLINE];
	} fail[HTTP_FAIL_EXPIRES] = {
		{ "Expires:\r\n   ", "Expires:\r\n   " },
		{ "Expires:", "Expires:" }
	};

	for (i = 0; i < HTTP_PASS_EXPIRES; i++) {
		fail_unless(http_parse(pass[i].line, 
				strlen(pass[i].line), &h), 
			query2str("fail: expires", pass[i].save));
		fail_unless(HTTP_EXPIRES == h.type,
			query2str("fail: expires", pass[i].save));
		fail_unless(0 == strcmp(h.data.expires, pass[i].expect),
			query2str("fail: expires", pass[i].save));
	}
	for (i = 0; i < HTTP_FAIL_EXPIRES; i++)
		fail_unless( ! http_parse(fail[i].line, 
					strlen(fail[i].line), &h), 
			query2str("fail-ok: expires", fail[i].save));
} END_TEST

START_TEST(http_parse_general)
{
#define HTTP_PASS_GENERAL (4)
#define HTTP_FAIL_GENERAL (2)
	struct http_header h;
	int 		i;

	struct {
		char line[HTTP_PARSE_MAXLINE];
		char save[HTTP_PARSE_MAXLINE];
		char key[HTTP_PARSE_MAXLINE];
		char val[HTTP_PARSE_MAXLINE];
	} pass[HTTP_PASS_GENERAL] = {
		{ "foo: bar", "foo: bar", "foo", "bar" },
		{ "foo:bar", "foo:bar", "foo", "bar" },
		{ "FOO:BAR", "FOO:BAR", "foo", "BAR" },
		{ "FOO:   \r\nBAR\r\n", "FOO:   \r\nBAR\r\n",
		  "foo", "BAR" },
	};

	struct {
		char line[HTTP_PARSE_MAXLINE];
		char save[HTTP_PARSE_MAXLINE];
	} fail[HTTP_FAIL_GENERAL] = {
		{ "foo", "foo" },
		{ "foo   \r\n", "foo   \r\n" },
	};

	for (i = 0; i < HTTP_PASS_GENERAL; i++) {
		fail_unless(http_parse(pass[i].line, 
				strlen(pass[i].line), &h), 
			query2str("fail: general", pass[i].save));
		fail_unless(0 == strcmp(pass[i].key, h.key),
			query2str("fail: general", pass[i].save));
		fail_unless(0 == strcmp(pass[i].val, h.val),
			query2str("fail: general", pass[i].save));
	}
	for (i = 0; i < HTTP_FAIL_GENERAL; i++)
		fail_unless( ! http_parse(fail[i].line, 
					strlen(fail[i].line), &h), 
			query2str("fail-ok: general", fail[i].save));
} END_TEST

int 
main(void)
{
	int 		nf;
	Suite 		*base;
	SRunner 	*sr;
	TCase		*tc;

	base = suite_create("base");

	tc = tcase_create("http");
	tcase_add_test(tc, http_parse_general);
	tcase_add_test(tc, http_parse_expires);
	tcase_add_test(tc, http_parse_pragma);
	tcase_add_test(tc, http_parse_content_type);
	tcase_add_test(tc, http_parse_content_language);
	tcase_add_test(tc, http_parse_location);
	tcase_add_test(tc, http_status_general);
	suite_add_tcase(base, tc);

	tc = tcase_create("uri");
	tcase_add_test(tc, uri_abs);
	suite_add_tcase(base, tc);

	sr = srunner_create(base);

	srunner_run_all(sr, CK_NORMAL);
	nf = srunner_ntests_failed(sr);
	srunner_free(sr);

#if CHECK_MAJOR_VERSION == 0 && CHECK_MINOR_VERSION < 9
	suite_free(base);
#endif

	return(nf == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
}
