/* $Id: up.c,v 1.35 2007-05-25 10:31:03 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>

#include <assert.h>
#include <err.h>
#include <getopt.h>
#include <inttypes.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "nlpcrawl.h"


/* TODO: more intelligently handle per-state codes. */
/* TODO: have long_opts for --sig-next-dead and so on (why not?). */


/* Default SCAN -> DIRTY expiration: 10 days. */
#define UP_DEF_TTEXP	(60 * 60 * 24 * 10)

/* Default DEAD -> (delete) time: 5 days. */
#define UP_DEF_TTDEL	(60 * 60 * 24 * 5)

/* Default UNREACH -> DIRTY time: 5 days. */
#define UP_DEF_TTREV	(60 * 60 * 24 * 5)

/* Default UNREACH -> DEAD interval: 5 intervals. */
#define UP_DEF_RTDEAD	(5)


static struct opt_t {
	uint64_t	ttrev;		/* Time-to-revival (or 0).   */
	uint64_t	ttdel;		/* Time-to-deletion (or 0).  */
	uint64_t	rtdead;		/* Revivals-to-dea (or 0)d.  */
	uint64_t	ttexp;		/* Time-to-expiration (or 0).*/
	int64_t		window;
	void		*cookie;	/* Active scan cookie.       */
	void		*q_del;		/* Write-queue for deletion. */
	void		*q_mod;		/* Write-queue for mods.     */
	char		*cache;		/* Cache-file directory.     */
} opts;


/*
 * Process SCAN-state entries for possible expiration.  Return 1 on
 * state-mod, 0 on no-mod.
 */
static int
proc_scan(struct nlpcdbrp *pair)
{
	assert(NULL != pair);
	assert(NLPCDBR_SCAN == pair->record.state);
	if (0 == opts.ttexp || UINT64_MAX == pair->record.expires) {
		dbg("modify: immediate expiration: %s",
				uri_topstring(&pair->record.uri));
		pair->record.state = NLPCDBR_DIRTY;
		return(1);
	}
	if (0 != pair->record.expires && pair->record.expires <
			pair->record.atime) {
		dbg("modify: expire: %" PRIu64 ", %" PRIu64 ": %s",
				pair->record.expires, 
				pair->record.atime,
				uri_topstring(&pair->record.uri));
		pair->record.state = NLPCDBR_DIRTY;
		return(1);
	}
	if (pair->record.mtime + opts.ttexp < pair->record.atime) {
		dbg("modify: expire (%" PRIu64 "): %" PRIu64 ": %s",
				opts.ttexp, pair->record.mtime,
				uri_topstring(&pair->record.uri));
		pair->record.state = NLPCDBR_DIRTY;
		return(1);
	}
	return(0);
}


/*
 * Process UNREACH-state entries for possible expiration.  Return 1 on
 * state-mod, 0 on no-mod.
 */
static int
proc_unreach(struct nlpcdbrp *pair)
{
	assert(NULL != pair);
	assert(NLPCDBR_UNREACH == pair->record.state);
	if (0 == opts.rtdead || opts.rtdead < pair->record.revived) {
		dbg("modify: revival exceeded (%" PRIu64 "): "
				"%" PRIu64 ", %s", opts.rtdead,
				pair->record.revived,
				uri_topstring(&pair->record.uri));
		pair->record.state = NLPCDBR_DEAD;
		return(1);
	}
	if (0 == opts.ttrev || pair->record.mtime + 
			opts.ttrev < pair->record.atime) {
		pair->record.revived++;
		dbg("modify: ttrev exceeded (%" PRIu64 "): "
				"%" PRIu64 ", %s", opts.ttrev,
				pair->record.revived,
				uri_topstring(&pair->record.uri));
		pair->record.state = NLPCDBR_DIRTY;
		return(1);
	}
	return(0);
}


/*
 * Process DEAD-state entries for possible expiration.  Return -1 on
 * deletion, 0 on no-mod.
 */
static int
proc_dead(struct nlpcdbrp *pair)
{
	assert(NULL != pair);
	assert(NLPCDBR_DEAD == pair->record.state);
	if (0 == opts.ttdel || pair->record.mtime + 
			opts.ttdel < pair->record.atime) {
		dbg("delete: ttdel exceeded (%" PRIu64 "): %s",
				opts.ttdel, 
				uri_topstring(&pair->record.uri));
		return(-1);
	}
	return(0);
}


/* 
 * Return -1 for deletion, 0 for no-state-change update, 1 for
 * state-change update. 
 */
static int
proc(struct nlpcdbrp *pair)
{
	assert(pair != NULL);
	if (NLPCDBR_DEAD == pair->record.state)
		return(proc_dead(pair));
	else if (NLPCDBR_SCAN == pair->record.state)
		return(proc_scan(pair));
	return(proc_unreach(pair));
}


static int
scan_del(void)
{
	struct uri	*addr;
	struct nlpcdbrp	*n;
	char		*path;
	int		flags;

	if (q_empty(&opts.q_del))
		return(0);

	while (NULL != (n = (struct nlpcdbrp *)q_rem(&opts.q_del))) {
		addr = &n->record.uri;
		path = nlpcdbr_hash2path(opts.cache, n->hash);
		if (SIGNAL_EXIT & (flags |= signal_test()))
			break;
		if ( ! db_del(n))
			diex("fail-delete: %s", uri_topstring(addr));
		if (-1 == remove(path)) 
			die("fail: remove: %s", path);
		free(n);
	}
	return(flags);
}


static int
scan_mod(void)
{
	int		mod, flags;
	struct uri	*addr;
	struct nlpcdbrp	*n;

	if (q_empty(&opts.q_mod))
		return(0);

	for (mod = 0, flags = 0; ; ) {
		n = (struct nlpcdbrp *)q_rem(&opts.q_mod);
		if (NULL == n)
			break;
		if (SIGNAL_EXIT & (flags |= signal_test()))
			break;
		addr = &n->record.uri;
		if ( ! db_put(n))
			diex("fail-modify: %s", uri_topstring(addr));
		if (NLPCDBR_DIRTY == n->record.type)
			mod++;
		free(n);
	}
	if (mod > 0)
		signal_next(NLPCDBR_DIRTY);
	return(flags);
}


static int
scan(void)
{
	struct db_query	query;
	int		rc, flags;
	struct nlpcdbrp pair, *n;
	void		*cookie;

	memset(&query, 0, sizeof(struct db_query));
	query.field = DB_FIELD_STATE;
	query.state = NLPCDBR_DEAD | NLPCDBR_UNREACH | NLPCDBR_SCAN;
	query.window = -1;

	db_window_fill(&query);
	db_query_begin(&cookie);

	flags = 0;

	for (;;) {
		if (1 != db_query_next(&cookie, &query, &pair))
			break;
		if (SIGNAL_EXIT & (flags |= signal_test()))
			break;
		pair.record.atime = time_uint64();
		if (0 != (rc = proc(&pair)))
			pair.record.mtime = pair.record.atime;

		calloc_s(n, 1, sizeof(struct nlpcdbrp));
		memcpy(n, &pair, sizeof(struct nlpcdbrp));

		if (-1 == rc) 
			q_add(&opts.q_del, n);
		else 
			q_add(&opts.q_mod, n);
	}

	db_query_close(&cookie);

	if (SIGNAL_EXIT & (flags |= scan_mod()))
		return(flags);
	if (SIGNAL_EXIT & (flags |= scan_del()))
		return(flags);
	return(flags);
}


int
main(int argc, char *argv[])
{
	int 		c, t;
	char		*dir;

	extern char	*__progname;
	extern char 	*optarg;
	extern int 	optind;

	static struct option longopts[] = {
		{ "ttrev", required_argument, NULL, 2 },
		{ "ttdel", required_argument, NULL, 3 },
		{ "ttexp", required_argument, NULL, 4 },
		{ "rtdead", required_argument, NULL, 5 },
		{ "force-rev", required_argument, NULL, 6 },
		{ "force-del", required_argument, NULL, 7 },
		{ "force-dead", required_argument, NULL, 8 },
		{ "force-exp", required_argument, NULL, 9 },
		{ NULL, 0, NULL, 0 }
	};

	dir = NLPC_DEFAULT_DIR;

	memset(&opts, 0, sizeof(struct opt_t));
	opts.ttdel = UP_DEF_TTDEL;
	opts.ttrev = UP_DEF_TTREV;
	opts.rtdead = UP_DEF_RTDEAD;
	opts.ttexp = UP_DEF_TTEXP;
	opts.cache = NLPC_DEFAULT_FILES;

	q_new(&opts.q_del);
	q_new(&opts.q_mod);

	while (-1 != (c = getopt_long(argc, argv, "hvdp:f:",
					longopts, NULL)))
		switch (c) {
		case (2):
			if ( ! parse_int(optarg, &t, 1, INT_MAX))
				diex("failed parse: --ttrev %s", 
						optarg);
			opts.ttrev = (uint64_t)t;
			break;
		case (3):
			if ( ! parse_int(optarg, &t, 1, INT_MAX))
				diex("failed parse: --ttdel %s", 
						optarg);
			opts.ttdel = (uint64_t)t;
			break;
		case (4):
			if ( ! parse_int(optarg, &t, 1, INT_MAX))
				diex("failed parse: --ttexp %s", 
						optarg);
			opts.ttexp = (uint64_t)t;
			break;
		case (5):
			if ( ! parse_int(optarg, &t, 1, INT_MAX))
				diex("failed parse: --rtdead %s", 
						optarg);
			opts.rtdead = (uint64_t)t;
			break;
		case (6):
			if (0 == opts.rtdead) 
				diex("cannot specify both --force-rev "
						"and --force-dead");
			opts.ttrev = 0;
			break;
		case (7):
			opts.ttdel = 0;
			break;
		case (8):
			if (0 == opts.ttrev) 
				diex("cannot specify both --force-rev "
						"and --force-dead");
			opts.rtdead = 0;
			break;
		case (9):
			opts.ttexp = 0;
			break;
		case ('d'):
			dbg_lvl++;
			break;
		case ('f'):
			opts.cache = optarg;
			break;
		case ('p'):
			dir = optarg;
			break;
		case ('v'):
			fprintf(stdout, "%s version %s\n", 
					__progname, VERSION);
			return(EXIT_SUCCESS);
		case ('h'):
			/*FALLTHROUGH*/
		default:
			fprintf(stderr, "usage: %s [-dhv] [-f path] "
					"[-p path] [LONG_OPTS]\n", 
					__progname);
			return(EXIT_FAILURE);
		}
	argc -= optind;
	argv += optind;

	signal_init(dir, NLPCDBR_UNREACH | 
			NLPCDBR_DEAD | NLPCDBR_SCAN);
	db_init(dir, 0);

	for (;;) {
		if (SIGNAL_EXIT & scan())
			break;
		if (SIGNAL_EXIT & signal_wait(0))
			break;
	}

	q_free(&opts.q_mod);
	q_free(&opts.q_del);
	db_teardown();
	signal_teardown();
	return(EXIT_SUCCESS);
}

